"""A representation of species' defaults."""

import gzip
import inspect
from collections.abc import Sequence
from dataclasses import asdict, dataclass, field, fields
from pathlib import Path

from pyfhiaims.control.chunk import AimsControlChunk
from pyfhiaims.errors import InvalidSpeciesInput
from pyfhiaims.species_defaults.basis_function import BASIS_TYPES
from pyfhiaims.species_defaults.basis_set import BasisSet
from pyfhiaims.species_defaults.electronic_configuration import (
    IonicElectronicConfiguration,
    ValenceElectronicConfiguration,
)
from pyfhiaims.species_defaults.integration_grid import IntegrationGrid


def replace_last(line, _):
    """Get the last term on a line."""
    return line.split()[-1]


def replace_last_float(line, _):
    """Get the last float on a line."""
    return float(line.split()[-1].replace("d", "e"))


def replace_last_bool(line, _):
    """Return True if option is True."""
    return line.split()[-1] == ".true."


def replace_last_int(line, _):
    """Get the last int on a line."""
    return int(line.split()[-1])


def replace_last_float_list(line, _):
    """Return the list of floats in a line."""
    return tuple([float(ll.replace("d", "e")) for ll in line.split()[1:]])


def append_div(line, val):
    """Append new values to a division."""
    # Check this doc string
    return [*val, (float(line.split()[1].replace("d", "e")), int(line.split()[2]))]


def append_inact_div(line, val):
    """Append a new value with an inactive div."""
    # Check this doc string
    return [*val, (float(line.split()[2].replace("d", "e")), int(line.split()[3]))]


SPECIES_KEYWORD_MAP = {
    "species": ("label", replace_last),
    "mass": ("mass", replace_last_float),
    "nucleus": ("nucleus", replace_last_float),
    "l_hartree": ("l_hartree", replace_last_int),
    "cut_pot": ("cut_pot", replace_last_float_list),
    "basis_dep_cutoff": ("basis_dep_cutoff", replace_last_float),
    "basis_acc": ("basis_acc", replace_last_float),
    "cite_reference": ("cite_reference", replace_last),
    "core": ("core", lambda line, _: (int(line.split()[1]), line.split()[2])),
    "core_states": ("core_states", replace_last_int),
    "cut_atomic_basis": ("cut_atomic_basis", replace_last_bool),
    "cut_core": (
        "cut_core",
        lambda line, _: (line.split()[1], float(line.split()[2].replace("d", "e"))),
    ),
    "cut_free_atom": (
        "cut_free_atom",
        lambda line, _: (line.split()[1], float(line.split()[2].replace("d", "e"))),
    ),
    "cutoff_type": ("cutoff_type", replace_last),
    "hirshfeld_param": ("hirshfeld_param", replace_last_float_list),
    "hubbard_coefficient": ("hubbard_coefficient", replace_last_float_list),
    "include_min_basis": ("include_min_basis", replace_last_bool),
    "innermost_max": ("innermost_max", replace_last_int),
    "logarithmic": ("logarithmic", replace_last_float_list),
    "plus_u": (
        "plus_u",
        lambda line, val: (val or [])
        + [
            (
                int(line.split()[1]),
                line.split()[2],
                float(line.split()[3].replace("d", "e")),
            ),
        ],
    ),
    "plus_u_ramping": ("plus_u_ramping", replace_last_float),
    "prodbas_acc": ("prodbas_acc", replace_last_float),
    "max_n_prodbas": ("max_n_prodbas", replace_last_int),
    "max_l_prodbas": ("max_l_prodbas", replace_last_int),
    "pure_gauss": ("pure_gauss", replace_last_bool),
    "species_default_type": ("species_default_type", replace_last),
}

ALLOWED_CUTOFF_TYPES = [
    "exp(1_x)_(1-x)2",
    "junquera",
    "x2_(1-x2)",
    None,
]


VALID_INPUTS = {
    "cutoff_type": ALLOWED_CUTOFF_TYPES,
    "species_default_type": ["minimal+s", None],
    "cite_reference": ["NAO-VCC-2013", None],
    "angular_grids": ["specified", "auto"],
    "cut_free_atom": ["finite", "infinite", None],
}


@dataclass
class SpeciesDefaults:
    """Defines a species (basis set) for FHI-aims.

    Parameters
    ----------
    basis_set: BasisSet
        The basis set for the species
    # header: list[str]
    #     The header for the species file
    label: str
        The label of the species to use in the geometry.in file
    mass: float
        The nuclear mass of the species
    nucleus: float
        The charge of the species
    # division_active: Sequence[tuple[float, int]]
    #     For specified angular_grids, the number of angular points on all radial
    #     shells that are within radius, but not within another, smaller division.
    #     Restrictions: Meaningful only in a block immediately following an
    #                   angular_grids specified line.
    #     radius : Outer radius (in Å) of this division.
    #     points : Integer number of angular points requested in this division (see
    #              force_lebedev for possible values).
    # division_inactive: Sequence[tuple[float, int]]
    #     (Commented out in species file)
    #     For specified angular_grids, the number of angular points on all radial
    #     shells that are within radius, but not within another, smaller division.
    #     Restrictions: Meaningful only in a block immediately following an
    #                   angular_grids specified line.
    #     radius : Outer radius (in Å) of this division.
    #     points : Integer number of angular points requested in this division (see
    #              force_lebedev for possible values).
    l_hartree: int
        The highest angular momentum component used in the multipole expansion
        of δn(r) into δñat,lm(r) for the present species.
    cut_pot: Vector3D
        Specifies the numerical parameters for the general (default) confinement
        potential vc(r) for all basis functions of this species.
            onset: specifies the default onset radius of the cutoff potential,
                   in Å (vc(r)=0 for r < r_onset).
            width: specifies the radial width w of the cutoff potential,
                   in Å (vc(r)=∞ for r > r_onset + w).
            scale: is a scaling parameter to increase or decrease the numerical
                   value of vc
    basis_dep_cutoff: bool | float
        Basis function dependent adjustment of the confinement potential for
        this species. If not .false., the onset of the basis confining potential
        (see cut_pot) is adjusted separately for each basis function, such that
        the norm of this basis function outside r_onset is smaller that threshold.
        The maximum possible onset radius is still given by the value explicitly
        specified by cut_pot
    basis_acc: float
        Technical cutoff criterion for on-site orthonormalization of radial
        functions. If initial normalization is below this value it is omitted
    cite_reference: str
        Triggers the addition of a specific citation to the end of the FHI-aims
        standard output for a given run. Given a key
    core: tuple[int, str]
        Defines the top “core” shell of the species for this angular momentum.
            n: radial quantum number
            l: character specifying angular momentum (s, p, d, f, ....)
        (experimental)
    core_states: int
        Number of core states (2 * core_states = number of core electrons)
        0 means don't use. This will interact with core in the future (experimental)
    cut_atomic_basis: bool
        If True basis_dep_cutoff keyword also applies to atomic-type (minimal)
        radial functions
    cut_core: tuple[str, float]
        Deprecated
        Can be used to define a separate (tighter) onset of the cutoff potential for
        all core radial functions.
        type : A string, either finite or infinite. Default: finite .
        radius : A real number, in Å: Onset radius for the cutoff potential,
                 as defined in the cut_pot tag. Default: same as onset in cut_pot .
    cut_free_atom: tuple[str, float]
        Adds a cutoff potential to the initial, non-spinpolarized free-atom
        calculation that yields free-atom densities and potentials for many basic
        tasks.
        type: either finite or infinite
        radius: A real number, in Å: Onset radius for the cutoff potential, as defined
                in the cut_pot tag
    cutoff_type: str
        Specifies the functional form of the confinement potential associated
        with this species
    hirshfeld_param: tuple[float, float, float]
        To explicitly allow setting the parameters for the Tkatchenko-Scheffler
        van der Waals correction. Does not apply to the vdw_ts keyword.
        C6 alpha R0 (See FHI-aims manuel)
    hubbard_coefficient: tuple[float, float, float, float]]
        See FHI-aims manuel, experimental feature for DFT + U
    include_min_basis: bool
        If False exclude the minimal basis of numerically tabulated free-atom
        basis functions (core and valence) from the basis set
    innermost_max: int
        If, after on-site orthonormalization, a radial function’s innermost extremum
        is inside the radial grid shell number, counting from the nucleus, that radial
        function is rejected in order to prevent inaccurate integrations.
    logarithmic: tuple[float, float, float]]
        Defines the dense one-dimensional “logarithmic” grid for the direct solution
        of all radial equations (free atom quantities, Hartree potential).
        r_min: is a real number (in bohr); the innermost point of the logarithmic grid
               is defined as r(1)=r_min/Z, where Z is the atomic number of the nucleus
               of the species . Default: 0.0001 bohr.
        r_max: is a real number (in bohr), the outermost point of the logarithmic grid,
               r(N). Default: 100 bohr.
        increment: is a real number, the increment factor α between successive
                   grid points, r(i) = α · r(i − 1). Default: 1.0123.
    # outer_grid: int
    #     Specify the number of angular grid points used outside the outermost
    #     division radius
    plus_u: Sequence[tuple[int, str, float]]
        Experimental—only for DFT+U. Adds a +U term to one specific shell
        of this species.
            n: the (integer) radial quantum number of the selected shell.
            l: is a character, specifying the angular momentum ( s, p, d, f, ...)
               of the selected shell.
            U: the value of the U parameter, specified in eV.
    prodbas_acc: float
        Technical cutoff criterion for on-site orthonormalization of auxiliary radial
        functions.
    max_l_prodbas: int
        Specifies the maximal angular quantum number for the auxiliary (product) basis
        function. Any possible auxiliary basis with an angular momentum higher than
        max_l_prodbas is excluded.
    max_n_prodbas: int
        Specifies the maximal principal quantum number for the regular basis
        function to be included in the auxiliary (product) basis construction
    pure_gauss: bool
        If True any Gaussian basis functions for this species will be purely
        spherical Gaussians
    # radial_multiplier: int]
    #     Systematically increases the radial integration grid density.
    #     number specifying the number of added subdivisions per basic grid spacing.
    species_default_type: str]
        Defines the default type of basis set

    """

    label: str = None
    mass: float = None
    nucleus: float = None
    l_hartree: int = None
    integration_grid: IntegrationGrid = None
    valence_config: ValenceElectronicConfiguration = None
    ionic_config: IonicElectronicConfiguration = None
    basis_set: BasisSet = None

    basis_dep_cutoff: bool | float = 1e-4
    cut_pot: tuple[float, float, float] = field(default_factory=tuple)
    basis_acc: float = None
    cite_reference: str = None
    core: tuple[int, str] = None
    core_states: int = None
    cut_atomic_basis: bool = None
    cut_core: tuple[str, float] = None
    cut_free_atom: tuple[str, float] = None
    cutoff_type: str = None
    hirshfeld_param: tuple[float, float, float] = None
    hubbard_coefficient: tuple[float, float, float, float] = None
    include_min_basis: bool = None
    innermost_max: int = None
    logarithmic: tuple[float, float, float] = None
    plus_u: Sequence[tuple[int, str, float]] = None
    prodbas_acc: float = None
    max_l_prodbas: int = None
    max_n_prodbas: int = None
    pure_gauss: bool = None
    species_default_type: str = None

    header: str | None = None

    def __post_init__(self):
        """Verify all inputs passed via init."""
        message = ""
        for key, val in asdict(self).items():
            if (
                key in VALID_INPUTS
                and (val if not isinstance(val, Sequence) else val[0])
                not in VALID_INPUTS[key]
            ):
                message += (
                    f"The input for key {key} ({val}) is invalid. "
                    f"Valid options are {VALID_INPUTS[key]}\n"
                )
        if len(message) != 0:
            raise InvalidSpeciesInput(message)

    def __str__(self):
        return self.content

    @property
    def content(self):
        """Get the text block for the species."""
        dct = asdict(self)
        for key in (
            "header",
            "basis_set",
            "integration_grid",
            "valence_config",
            "ionic_config",
        ):
            dct.pop(key)

        file_contents = [
            f"{self.header}",
            f"  species               {dct.pop('label')}",
            "#     global species definitions",
            f"  nucleus               {dct.pop('nucleus')}",
            f"  mass                  {dct.pop('mass')}",
            "#",
            f"  l_hartree             {dct.pop('l_hartree')}",
            "#",
        ]
        cut_pot = dct.pop("cut_pot")
        file_contents += [
            f"  cut_pot               {cut_pot[0]} {cut_pot[1]} {cut_pot[2]}",
            f"  basis_dep_cutoff      {dct.pop('basis_dep_cutoff')}",
            "#",
            f"{self.integration_grid}",
        ]

        if any(val is not None for val in dct.values()):
            file_contents += [
                "#" * 82,
                "#",
                "# Optional basis settings passed to pyfhiaims",
                "#",
                "#" * 82,
            ]
            for key, val in dct.items():
                if val is None:
                    continue

                if key == "plus_u":
                    for pu in val:
                        file_contents.append(
                            f"    {key:<10s}{' '.join([str(v) for v in pu])}"
                        )
                elif isinstance(val, tuple):
                    file_contents.append(
                        f"  {key:<22s}{' '.join([str(v) for v in val])}"
                    )
                else:
                    file_contents.append(f"  {key:<22s}{val}")

        file_contents += [
            "#" * 82,
            "#",
            '#  Definition of "minimal" basis',
            "#",
            "#" * 82,
            "#     valence basis states",
            f"{self.valence_config}",
            "#     ion occupancy",
            f"{self.ionic_config}",
        ]

        file_contents += [
            "#" * 82,
            "#",
            "#  Suggested additional basis functions. For production calculations,",
            "#  uncomment them one after another "
            "(the most important basis functions are",
            "#  listed first).",
            "#",
            "#  These were set using pyfhiaims, "
            "original may files have additional comments",
            "#",
            "#" * 82,
        ]

        file_contents.append(f"{self.basis_set}")
        return "\n".join(file_contents)

    @classmethod
    def from_file(cls, filename: str | Path) -> "SpeciesDefaults":
        """Create a species from an FHI-aims species file."""
        lines = None
        for fname in [filename, f"{filename}.gz"]:
            if not Path(fname).exists():
                continue

            if fname[-3:] == ".gz":
                with gzip.open(fname, mode="rt") as in_file:
                    lines = in_file.readlines()
            else:
                with open(fname) as in_file:
                    lines = in_file.readlines()
        if lines is None:
            raise FileNotFoundError(f"Species file {filename} does not exit.")
        return cls.from_strings(lines)

    @classmethod
    def from_strings(cls, config_str: list[str]) -> "SpeciesDefaults":
        """Create a species from a list of lines from a file with species' defaults.

        The keywords from the species' defaults are organized in logical chunks to
        facilitate comparing between two different SpeciesDefaults objects.

        Parameters
        ----------
        cls:
            The class keyword
        config_str: list[str]
            a list of strings from the species' defaults.

        Returns
        -------
        SpecieseDefaults
            The object defined by the strings from the final

        """
        parameters = {}
        # get the keywords pertaining to chunk objects
        kw_map = {}
        # as well as class objects for chunks
        chunk_types = {}
        for f in fields(cls):
            if inspect.isclass(f.type) and issubclass(f.type, AimsControlChunk):
                for kw in f.type.keywords:
                    kw_map[kw] = f.name
                chunk_types[f.name] = f.type
                parameters[f.name] = []

        lines = [line.strip() for line in config_str]

        for il, line in enumerate(lines):
            if len(line) == 0:
                continue
            split_line = line.split()
            if (len(split_line) <= 1) and (split_line[0] == "#"):
                continue

            if split_line[0] != "#":
                key = split_line[0]
            else:
                key = f"{split_line[0]} {split_line[1]}"

            if key in SPECIES_KEYWORD_MAP:
                param_setter = SPECIES_KEYWORD_MAP[key]
                cur_val = parameters.get(param_setter[0], None)
                parameters[param_setter[0]] = param_setter[1](line, cur_val)
                continue
            # get chunk lines
            if key in kw_map:
                parameters[kw_map[key]].append(line)

            # basis functions appear below
            if key in BASIS_TYPES:
                parameters["basis_set"] = BasisSet.from_string("\n".join(lines[il:]))
                break
        for chunk, chunk_type in chunk_types.items():
            parameters[chunk] = chunk_type.from_strings(parameters[chunk])

        header_lines = []
        if "########" in lines[0]:
            header_lines.append(lines[0])
            il = 1
            while len(lines[il]) == 0 or lines[il][0] == "#":
                header_lines.append(lines[il])
                il += 1
            parameters["header"] = "\n".join(header_lines)
        else:
            parameters["header"] = "\n".join(
                [
                    "",
                    "#" * 82,
                    "#",
                    f"# Read in species defaults for {parameters['label']}"
                    f" atom from pyfhiaims",
                    "#",
                    "#" * 82,
                ]
            )

        return cls(**parameters)



"""Tests for electronic_configuration module."""

from textwrap import dedent

from pyfhiaims.species_defaults.electronic_configuration import (
    ElectronicConfiguration,
    ElectronicConfigurationType,
)

valence_str = dedent(
    """\
    valence      2  s   2.
    valence      2  p   3."""
)

ionic_str = dedent(
    """\
    ion_occ      6  s   1.
    ion_occ      6  p   1.
    ion_occ      5  d  10.
    ion_occ      4  f  14."""
)


def test_electronic_configuration():
    """Test electronic_configuration"""
    el_config = ElectronicConfiguration.from_strings(valence_str.split("\n"))
    assert el_config.type == ElectronicConfigurationType.VALENCE
    assert el_config.n_electrons == 7
    assert dedent(str(el_config)) == valence_str

    el_config = ElectronicConfiguration.from_strings(ionic_str.split("\n"))
    assert el_config.type == ElectronicConfigurationType.IONIC
    assert el_config.n_electrons == 80
    assert dedent(str(el_config)) == ionic_str



"""Defines class/functions to write input and parse output for FHI-aims."""

import os
import time
import warnings
from collections.abc import Iterable
from pathlib import Path
from typing import Any, TextIO

import numpy as np
from ase import Atoms
from ase.calculators.calculator import kpts2mp
from ase.calculators.singlepoint import SinglePointDFTCalculator
from ase.units import Ang, fs
from ase.utils import deprecated, reader, writer

from pyfhiaims.control.control import AimsControl
from pyfhiaims.geometry.atom import FHIAimsAtom
from pyfhiaims.geometry.geometry import AimsGeometry
from pyfhiaims.outputs.stdout import AimsParseError, AimsStdout

v_unit = Ang / (1000.0 * fs)

LINE_NOT_FOUND = object()


# Read aims geometry files
@reader
def read_aims(fd: TextIO | str | Path, apply_constraints=True) -> Atoms:
    """
    Import FHI-aims geometry type files.

    Reads unitcell, atom positions and constraints from
    a geometry.in file.
    """
    geometry = AimsGeometry.from_file(fd)
    atoms = geometry.ase_atoms

    if apply_constraints:
        atoms.set_positions(atoms.get_positions())
    return atoms


def get_aims_header() -> str:
    """Returns the header for aims input files"""
    lines = ["#" + "=" * 79]
    for line in [
        "Created using the Atomic Simulation Environment (ASE)",
        time.asctime(),
    ]:
        lines.append("# " + line)
    return "\n".join(lines)


def _write_velocities_alias(args: list, kwargs: dict[str, Any]) -> bool:
    arg_position = 5
    if len(args) > arg_position and args[arg_position]:
        args[arg_position - 1] = True
    elif kwargs.get("velocities", False):
        if len(args) < arg_position:
            kwargs["write_velocities"] = True
        else:
            args[arg_position - 1] = True
    else:
        return False
    return True


# Write aims geometry files
@deprecated(
    "Use of `velocities` is deprecated, please use `write_velocities`",
    category=FutureWarning,
    callback=_write_velocities_alias,
)
@writer
def write_aims(
    fd: TextIO | str | Path,
    atoms: Atoms,
    scaled: bool = False,
    geo_constrain: bool = False,
    write_velocities: bool = False,
    velocities: bool = False,
    ghosts: None | Iterable[int] = None,
    info_str: None | str = None,
    wrap: bool = False,
):
    """
    Method to write FHI-aims geometry files.

    Writes the atoms positions and constraints (only FixAtoms is
    supported at the moment).

    Args:
        fd: TextIO | str | Path
            File to output structure to
        atoms: Atoms
            structure to output to the file
        scaled: bool
            If True use fractional coordinates instead of Cartesian coordinates
        geo_constrain: bool
            If true add parameteric constraints
        write_velocities: bool
            If True add the atomic velocity vectors to the file
        velocities: bool
            NOT AN ARRAY OF VELOCITIES, but the legacy version of
            `write_velocities`
        ghosts: list[int]
            A list of indexes, 1 = ghost, regular atom otherwise
        info_str: str
            A string to be added to the header of the file
        wrap: bool
            Wrap atom positions to cell before writing

    .. deprecated:: 3.23.0
        Use of ``velocities`` is deprecated, please use ``write_velocities``.

    """
    make_empty = None
    if ghosts is not None:
        make_empty = [ii for ii, ind in enumerate(ghosts) if ind == 1]

    geometry = AimsGeometry.from_atoms(
        atoms,
        remove_velocities=not write_velocities,
        geo_constrain=geo_constrain,
        make_empty=make_empty,
        scaled=scaled,
        wrap=wrap,
    )

    header = get_aims_header()
    # If writing additional information is requested via info_str:
    if info_str is not None:
        header += "\n# Additional information:\n"
        if isinstance(info_str, list):
            header += "\n".join([f"#  {s}" for s in info_str])
        else:
            header += f"# {info_str}"
        header += "\n"

    header += "#=======================================================\n"
    geometry.write_file(fd, header)


def get_species_directory(species_dir: str | Path | None = None):
    """Get the directory where the basis set information is stored

    If the requested directory does not exist then raise an Error

    Parameters
    ----------
    species_dir: str
        Requested directory to find the basis set info from. E.g.
        `~/aims2022/FHIaims/species_defaults/defaults_2020/light`.

    Returns
    -------
    Path
        The Path to the requested or default species directory.

    Raises
    ------
    RuntimeError
        If both the requested directory and the default one is not defined
        or does not exit.
    """
    if species_dir is None:
        species_dir = os.environ.get("AIMS_SPECIES_DIR")

    if species_dir is None:
        raise RuntimeError(
            "Missing species directory!  Use species_dir "
            + "parameter or set $AIMS_SPECIES_DIR environment variable."
        )

    species_path = Path(species_dir)
    if not species_path.exists():
        raise RuntimeError(f"The requested species_dir {species_dir} does not exist")

    return species_path


# Write aims control.in files
@writer
def write_control(
    fd: TextIO | str | Path,
    atoms: Atoms,
    parameters: dict[str, Any],
    verbose_header: bool = False,
):
    """
    Write the control.in file for FHI-aims

    Parameters
    ----------
    fd: TextIO | str | Path
        The file object to write to
    atoms: Atoms
        The Atoms object for the requested calculation
    parameters: dict[str, Any]
        The dictionary of all paramters for the calculation
    verbose_header: bool
        If True then explcitly list the paramters used to generate the
        control.in file inside the header

    """
    parameters = dict(parameters)

    if parameters["xc"] == "LDA":
        parameters["xc"] = "pw-lda"

    if "kpts" in parameters:
        mp = kpts2mp(atoms, parameters.pop("kpts"))
        dk = 0.5 - 0.5 / np.array(mp)
        parameters["k_grid"] = tuple(mp)
        parameters["k_offset"] = tuple(dk)

    parameters["species_dir"] = get_species_directory(parameters.pop("species_dir"))

    control_in = AimsControl(
        parameters=parameters,
        outputs=parameters.pop("output"),
    )

    geometry = AimsGeometry.from_atoms(atoms)

    control_in.write_file(
        geometry=geometry,
        writer=fd,
        header=get_aims_header(),
        verbose_header=verbose_header,
    )


@reader
def read_aims_output(
    fd: TextIO | str | Path, index: int | slice = -1, non_convergence_ok: bool = False
) -> Atoms | list[Atoms]:
    """
    Import FHI-aims output files with all data available

    Parameters
    ----------
    fd: TextIO | str | Path
        The file object to write to
    index: slice | int
        The images to return
    non_convergence_ok: bool
        True if a non-converged result is okay

    Returns
    -------
    Atoms | list[Atoms]
        The requested Atoms objects
    """
    output = AimsStdout(fd)

    if isinstance(index, int):
        loop_inds = [index]
    else:
        loop_inds = range(output.n_images)
        loop_inds = loop_inds[index]

    atoms_list = []
    for ind in loop_inds:
        image = output[ind]
        if not non_convergence_ok and (not image.converged):
            raise AimsParseError("The calculation did not converge properly.")
        atoms = image.geometry.ase_atoms
        atoms.calc = SinglePointDFTCalculator(
            atoms,
            energy=image.free_energy,  # TARP: This is the force-consistent energy
            free_energy=image.free_energy,
            forces=image.forces,
            stress=image.stress,
            stresses=image.stresses,
            magmom=image.magmom,
            dipole=image.dipole,
            dielectric_tensor=image.dielectric_tensor,
            polarization=image.polarization,
        )
        atoms_list.append(atoms)
    if isinstance(index, int):
        return atoms_list[0]

    return atoms_list


def read_aims_results(
    fd: TextIO | str | Path,
    index: int | slice = -1,
    non_convergence_ok: bool = False,
    append_metadata: bool = False,
    verbosity: str = "converged",
) -> dict[str, Any] | list[dict[str, Any]]:
    """
    Import FHI-aims output files with all data available as a dict

    Parameters
    ----------
    fd: TextIO | str | Path
        The file object to write to
    index: slice | int
        The images to return
    non_convergence_ok: bool
        True if a non-converged result is okay
    append_metadata: bool
        True add non-property based keys to the results dictionary
    verbosity: str
        Results verbosity level, can be "single_point" or "all"
    Returns
    -------
    dict[str, Any] | list[dict[str, Any]]
        The requested results Dictionaries
    """
    output = AimsStdout(fd)

    if isinstance(index, int):
        loop_inds = [index]
    else:
        loop_inds = range(output.n_images)
        loop_inds = loop_inds[index]

    image_dicts: list[dict[str, Any]] = []
    for ind in loop_inds:
        image = output[ind]
        if not non_convergence_ok and (not image.converged):
            raise AimsParseError("The calculation did not converge properly.")
        image_dicts.append(image.get_results(verbosity=verbosity))

    results: dict[str, Any] | list[dict[str, Any]] = {}
    if append_metadata:
        results = {
            "calc_metadata": {
                key: output.results[key]
                for key in output.keys()
                if key
                not in [
                    "final",
                    "ionic_steps",
                    "num_scf_steps",
                    "num_relax_steps",
                    "num_ionic_steps",
                    *output.results["final"].keys(),
                ]
            },
            "images": image_dicts,
        }
    else:
        results = image_dicts

    if isinstance(index, int) and not append_metadata:
        return image_dicts[0]

    return results



"""Tests for `utils.typecast` submodule."""

import numpy as np

from pyfhiaims.utils.typecast import to_matrix3d, to_vector3d


def test_typecast():
    """Tests utils.typecast submodule."""
    a = (3, 2, 3)
    vec = to_vector3d(a)
    assert isinstance(vec, np.ndarray)
    assert vec.shape == (3,)
    assert vec.dtype == float
    b = [[True, False, True], [False, True, False], [True, True, False]]
    mat = to_matrix3d(b, dtype=bool)
    assert isinstance(mat, np.ndarray)
    assert mat.shape == (3, 3)
    assert mat.dtype == bool



"""Type converters for parsed values."""

import datetime
import re

from pyfhiaims import AimsGeometry


def _to_builtin(pattern, builtin_type):
    """Convert pattern to the builtin type."""

    def _helper(line, _):
        match = re.search(pattern, line, re.DOTALL | re.MULTILINE)
        if match:
            return builtin_type(match.group(1))
        return None

    return _helper


def to_atoms(pattern, from_input=False):
    """Convert aims.out geometry representation to AimsGeometry."""
    from pyfhiaims.geometry.geometry import AimsGeometry

    def _helper(line, _):
        # Capture everything to the next dashed line including it
        match = re.search(
            rf"{pattern}.*\n([\s\S]*?)\n *-{{60,}}", line, re.MULTILINE
        )
        if match is None:
            return None
        lines = match.group()
        # remove fractional coordinates
        lines = lines[: lines.find("Fractional coordinates")]
        lines = lines.split("\n")[2:-1]
        return AimsGeometry.from_strings(lines)
        # with tempfile.TemporaryDirectory() as tmp:
        #     file_name = Path(tmp) / "geometry.in"
        #     with open(file_name, "w") as f:
        #         f.write("\n".join(lines))
        #     atoms = read(file_name, format="aims")

    return _helper

def bsse_to_atoms():
    """Convert BSSE geometry representation to AimsGeometry."""
    def _helper(line, _):
        match = re.search(
            r"-{59,}\n([\s\S]*?)\n *-{59,}", line, re.MULTILINE
        )
        if match is None:
            return None
        lines = match.group().split("\n")[2:-1]
        geom_in = []
        is_empty = False
        for s in lines:
            if "x [A]" in s:
                is_empty = "Ghost" in s
                continue
            bits = s.split()[3:]
            geom_in.append(f"{'empty' if is_empty else 'atom'} "
                           f"{' '.join(bits[1:4])} {bits[0]}")
        return AimsGeometry.from_strings(geom_in)
    return _helper


def to_bool(pattern):
    """Convert pattern to bool."""

    def _helper(line, _):
        return bool(re.search(pattern, line, re.DOTALL | re.MULTILINE))

    return _helper


def to_float(pattern):
    """Convert pattern to float."""
    return _to_builtin(pattern, float)


def to_int(pattern):
    """Convert pattern to int."""
    return _to_builtin(pattern, int)


def to_str(pattern):
    """Convert pattern to date."""
    return _to_builtin(pattern, str)


def to_date(pattern):
    """Convert pattern to date."""
    return _to_builtin(
        pattern, lambda x: datetime.datetime.strptime(x, "%Y%m%d").date()
    )


def to_time(pattern):
    """Convert pattern to a time."""
    return _to_builtin(
        pattern, lambda x: datetime.datetime.strptime(x, "%H%M%S.%f").time()
    )


def to_table(
    pattern,
    *,
    num_rows: int | str,
    header: int = 1,
    dtype: list[type] | tuple[type] | None = None,
):
    """Convert `num_rows` lines after `header` indicated by `pattern` to
    Python 2D list.
    """
    dtype = dtype if dtype is not None else []

    def _helper(line, metadata):
        # we need a local scope variable for not to lose access to num_rows
        n_rows = num_rows if isinstance(num_rows, int) else metadata[num_rows]
        match = re.search(
            rf"{pattern}\n((?:.*\n){{{header+n_rows-1}}})", line, re.MULTILINE
        )
        if match is None:
            return None
        table = match.group(1).split("\n")[header - 1 : -1]
        types = dtype + ([None] * (len(table[0].split()) - len(dtype)))
        result = []
        for table_line in table:
            result.append(
                [
                    t(v)
                    for t, v in zip(types, table_line.split(), strict=False)
                    if t is not None
                ]
            )
        # make it 1D list if only one column is asked for in the parser
        if len(result[0]) == 1:
            result = [r[0] for r in result]
        return result

    return _helper


def to_vector(pattern: str, *, dtype: type = float, multistring: bool = False):
    """Convert a set of numbers to a 1D numpy array.

    Parameters
    ----------
    pattern: str
        A regular expression pattern.
    dtype: type
        A Python type to convert to.
    multistring: bool
        If True, checks for different occurrences of the pattern in a line
        (different matches).

        If False, finds the numbers written in one line
        (different groups within one match).

    """

    def _helper(line, _):
        match = re.findall(pattern, line, re.DOTALL | re.MULTILINE)
        if not match:
            return None
        return (
            [dtype(x) for x in match] if multistring else [dtype(x) for x in match[0]]
        )

    return _helper


def to_matrix(
    pattern: str,
    *,
    dtype: list[type] | tuple[type] | None = None,
):
    """Convert a multi-line string matching a pattern into a 2D matrix.

    Parameters
    ----------
    pattern : str
        A regular expression pattern that captures rows of data as groups.
    dtype : list[type] | tuple[type] | None
        A list or tuple of Python types that specifies the type of each
        column. The length of `dtype` must match the number of groups in
        `pattern`. If None, no conversion is performed.

    Notes
    -----
    - Each match of the `pattern` corresponds to a row in the resulting 2D matrix.
    - The number of data columns in the pattern must match the number of dtypes
    specified.

    """

    def _helper(line, _):
        match = re.findall(pattern, line, re.DOTALL | re.MULTILINE)
        if not match:
            return None
        res = []
        for m in match:
            assert len(m) == len(dtype)
            res.append(
                [
                    dtype_i(m_i)
                    for dtype_i, m_i in zip(dtype, m, strict=False)
                    if dtype_i is not None
                ]
            )
        return res

    return _helper



"""Classes for reading/manipulating/writing FHI-aims cube files.

Works for aims cube objects
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import numpy as np
from monty.json import MontyDecoder, MSONable

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Self

    from pymatgen.util.typing import Tuple3Floats, Tuple3Ints

__author__ = "Thomas A. R. Purcell"
__version__ = "1.0"
__email__ = "purcellt@arizona.edu"
__date__ = "July 2024"

ALLOWED_AIMS_CUBE_TYPES = (
    "delta_density",
    "spin_density",
    "stm",
    "total_density",
    "total_density_integrable",
    "long_range_potential",
    "hartree_potential",
    "xc_potential",
    "delta_v",
    "ion_dens",
    "dielec_func",
    "elf",
)

ALLOWED_AIMS_CUBE_TYPES_STATE = (
    "first_order_density",
    "eigenstate",
    "eigenstate_imag",
    "eigenstate_density",
)

ALLOWED_AIMS_CUBE_FORMATS = (
    "cube",
    "gOpenMol",
    "xsf",
)


@dataclass
class AimsCube(MSONable):
    """The FHI-aims cubes.

    Attributes:
        type (str): The value to be outputted as a cube file
        origin (Sequence[float] or tuple[float, float, float]): The origin of the cube
        edges (Sequence[Sequence[float]]): Specifies the edges of a cube: dx, dy, dz
            dx (float): The length of the step in the x direction
            dy (float): The length of the step in the y direction
            dx (float): The length of the step in the x direction
        points (Sequence[int] or tuple[int, int, int]): The number of points
            along each edge
        spin_state (int): The spin-channel to use either 1 or 2
        kpoint (int): The k-point to use (the index of the list printed from
            `output k_point_list`)
        filename (str): The filename to use
        format (str): The format to output the cube file in: cube, gOpenMol, or xsf
        elf_type (int): The type of electron localization function to use (
            see FHI-aims manual)

    """

    type: str = field(default_factory=str)
    origin: Sequence[float] | Tuple3Floats = field(
        default_factory=lambda: [0.0, 0.0, 0.0]
    )
    edges: Sequence[Sequence[float]] = field(default_factory=lambda: 0.1 * np.eye(3))
    points: Sequence[int] | Tuple3Ints = field(default_factory=lambda: [0, 0, 0])
    format: str = "cube"
    spin_state: int | None = None
    kpoint: int | None = None
    filename: str | None = None
    elf_type: int | None = None

    def __eq__(self, other: object) -> bool:
        """Check if two cubes are equal to each other."""
        if not isinstance(other, AimsCube):
            return NotImplemented

        if self.type != other.type:
            return False

        if not np.allclose(self.origin, other.origin):
            return False

        if not np.allclose(self.edges, other.edges):
            return False

        if not np.allclose(self.points, other.points):
            return False

        if self.format != other.format:
            return False

        if self.spin_state != other.spin_state:
            return False

        if self.kpoint != other.kpoint:
            return False

        if self.filename != other.filename:
            return False

        return self.elf_type == other.elf_type

    def __hash__(self):
        """Return the hash of the AimsCube object."""
        return hash(self.to_json())

    def __post_init__(self) -> None:
        """Check the inputted variables to make sure they are correct.

        Raises:
            ValueError: If any of the inputs is invalid

        """
        split_type = self.type.split()
        cube_type = split_type[0]
        if split_type[0] in ALLOWED_AIMS_CUBE_TYPES:
            if len(split_type) > 1:
                raise ValueError(
                    f"{cube_type=} can not have a state associated with it"
                )
        elif split_type[0] in ALLOWED_AIMS_CUBE_TYPES_STATE:
            if len(split_type) != 2:
                raise ValueError(f"{cube_type=} must have a state associated with it")
        else:
            raise ValueError("Cube type undefined")

        if self.format not in ALLOWED_AIMS_CUBE_FORMATS:
            raise ValueError(
                f"{self.format} is invalid. Cube files must have a format of "
                f"{ALLOWED_AIMS_CUBE_FORMATS}"
            )

        valid_spins = (1, 2, None)
        if self.spin_state not in valid_spins:
            raise ValueError(f"Spin state must be one of {valid_spins}")

        if len(self.origin) != 3:
            raise ValueError("The cube origin must have 3 components")

        if len(self.points) != 3:
            raise ValueError("The number of points per edge must have 3 components")

        if len(self.edges) != 3:
            raise ValueError("Only three cube edges can be passed")

        for edge in self.edges:
            if len(edge) != 3:
                raise ValueError("Each cube edge must have 3 components")

        if self.elf_type is not None and self.type != "elf":
            raise ValueError(
                "elf_type is only used when the cube type is elf. "
                "Otherwise it must be None"
            )

    @property
    def control_block(self) -> str:
        """The block of text for the control.in file of the Cube."""
        cb = f"output cube {self.type}\n"
        cb += (
            f"    cube origin {self.origin[0]: .12e} {self.origin[1]: .12e} "
            f"{self.origin[2]: .12e}\n"
        )
        for idx in range(3):
            cb += f"    cube edge {self.points[idx]} "
            cb += f"{self.edges[idx][0]: .12e} "
            cb += f"{self.edges[idx][1]: .12e} "
            cb += f"{self.edges[idx][2]: .12e}\n"
        cb += f"    cube format {self.format}\n"
        if self.spin_state is not None:
            cb += f"    cube spinstate {self.spin_state}\n"
        if self.kpoint is not None:
            cb += f"    cube kpoint {self.kpoint}\n"
        if self.filename is not None:
            cb += f"    cube filename {self.filename}\n"
        if self.elf_type is not None:
            cb += f"    cube elf_type {self.elf_type}\n"

        return cb

    def as_dict(self) -> dict[str, Any]:
        """Get a dictionary representation of the geometry.in file."""
        dct: dict[str, Any] = {}
        dct["@module"] = type(self).__module__
        dct["@class"] = type(self).__name__
        dct["type"] = self.type
        dct["origin"] = self.origin
        dct["edges"] = self.edges
        dct["points"] = self.points
        dct["format"] = self.format
        dct["spin_state"] = self.spin_state
        dct["kpoint"] = self.kpoint
        dct["filename"] = self.filename
        dct["elf_type"] = self.elf_type
        return dct

    @classmethod
    def from_dict(cls, dct: dict[str, Any]) -> Self:
        """Initialize from dictionary.

        Args:
            dct (dict[str, Any]): The MontyEncoded dictionary

        Returns:
            AimsCube

        """
        attrs = (
            "type",
            "origin",
            "edges",
            "points",
            "format",
            "spin_state",
            "kpoint",
            "filename",
            "elf_type",
        )
        decoded = {key: MontyDecoder().process_decoded(dct[key]) for key in attrs}

        return cls(**decoded)



"""Module defining base FHI-aims input set and generator."""

from __future__ import annotations

import copy
import json
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any
from warnings import warn

import numpy as np
from monty.json import MontyDecoder, MontyEncoder
from pymatgen.core import Molecule, Structure
from pymatgen.io.aims.inputs import AimsControlIn, AimsGeometryIn
from pymatgen.io.aims.outputs import AimsOutput
from pymatgen.io.core import InputFile, InputGenerator, InputSet

from pyfhiaims.outputs.stdout import AimsParseError

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from pymatgen.util.typing import PathLike

TMPDIR_NAME: str = "tmpdir"
OUTPUT_FILE_NAME: str = "aims.out"
CONTROL_FILE_NAME: str = "control.in"
PARAMS_JSON_FILE_NAME: str = "parameters.json"
GEOMETRY_FILE_NAME: str = "geometry.in"


DEFAULT_AIMS_PROPERTIES = (
    "energy",
    "free_energy",
    "forces",
    "stress",
    "stresses",
    "dipole",
    "magmom",
)


class AimsInputSet(InputSet):
    """A class to represent a set of Aims inputs."""

    def __init__(
        self,
        parameters: dict[str, Any],
        structure: Structure | Molecule,
        properties: Sequence[str] = ("energy", "free_energy"),
    ) -> None:
        """Construct the AimsInputSet.

        Args:
            parameters (dict[str, Any]): The ASE parameters object for the calculation
            structure (Structure or Molecule): The Structure/Molecule objects to
                create the inputs for
            properties (Sequence[str]): The properties to calculate for the calculation
        """
        self._parameters = parameters
        self._structure = structure
        self._properties = properties

        aims_control_in, aims_geometry_in = self.get_input_files()
        super().__init__(
            inputs={
                CONTROL_FILE_NAME: aims_control_in,
                GEOMETRY_FILE_NAME: aims_geometry_in,
                PARAMS_JSON_FILE_NAME: json.dumps(self._parameters, cls=MontyEncoder),
            }
        )

    def get_input_files(self) -> tuple[str, str]:
        """Get the input file contents for the calculation.

        Returns:
            tuple[str, str]: The contents of the control.in and geometry.in file
        """
        property_flags = {
            "forces": "compute_forces",
            "stress": "compute_analytical_stress",
            "stresses": "compute_heat_flux",
        }
        updated_params = dict(**self._parameters)
        for prop in self._properties:
            aims_name = property_flags.get(prop)
            if aims_name is not None:
                updated_params[aims_name] = True

        aims_geometry_in = AimsGeometryIn.from_structure(self._structure)
        aims_control_in = AimsControlIn(updated_params)

        return (
            aims_control_in.get_content(structure=self._structure),
            aims_geometry_in.content,
        )

    @property
    def control_in(self) -> str | slice | InputFile:
        """The control.in file contents."""
        return self[CONTROL_FILE_NAME]

    @property
    def geometry_in(self) -> str | slice | InputFile:
        """The geometry.in file contents."""
        return self[GEOMETRY_FILE_NAME]

    @property
    def params_json(self) -> str | slice | InputFile:
        """The JSON representation of the parameters dict."""
        return self[PARAMS_JSON_FILE_NAME]

    def set_parameters(self, *args, **kwargs) -> dict[str, Any]:
        """Set the parameters object for the AimsTemplate.

        This sets the parameters object that is passed to an AimsTemplate and
        resets the control.in file

        One can pass a dictionary mapping the aims variables to their values or
        the aims variables as keyword arguments. A combination of the two
        options is also allowed.

        Returns:
            dict[str, Any]: dictionary with the variables that have been added.
        """
        self._parameters.clear()
        for arg in args:
            self._parameters.update(arg)

        self._parameters.update(kwargs)

        aims_control_in, _ = self.get_input_files()

        self.inputs[CONTROL_FILE_NAME] = aims_control_in
        self.inputs[PARAMS_JSON_FILE_NAME] = json.dumps(self._parameters, cls=MontyEncoder)

        inputs = {str(key): val for key, val in self.inputs.items()}
        self.__dict__.update(inputs)

        return self._parameters

    def remove_parameters(
        self,
        keys: Iterable[str] | str,
        strict: bool = True,
    ) -> dict[str, Any]:
        """Remove the aims parameters listed in keys.

        This removes the aims variables from the parameters object.

        Args:
            keys (Iterable[str] or str): string or list of strings with the names of
                the aims parameters to be removed.
            strict (bool): whether to raise a KeyError if one of the aims parameters
                to be removed is not present.

        Returns:
            dict[str, Any]: Dictionary with the variables that have been removed.
        """
        if isinstance(keys, str):
            keys = [keys]
        for key in keys:
            if key not in self._parameters:
                if strict:
                    raise ValueError(f"{key=} not in {list(self._parameters)=}")
                continue

            del self._parameters[key]

        return self.set_parameters(**self._parameters)

    def set_structure(self, structure: Structure | Molecule):
        """Set the structure object for this input set.

        Args:
            structure (Structure or Molecule): The new Structure or Molecule
                for the calculation
        """
        self._structure = structure

        aims_control_in, aims_geometry_in = self.get_input_files()
        self.inputs[GEOMETRY_FILE_NAME] = aims_geometry_in
        self.inputs[CONTROL_FILE_NAME] = aims_control_in
        inputs = {str(key): val for key, val in self.inputs.items()}
        self.__dict__.update(inputs)


@dataclass
class AimsInputGenerator(InputGenerator):
    """
    A class to generate Aims input sets.

    Attributes:
        user_params (dict[str, Any]): Updates the default
            parameters for the FHI-aims calculator
        user_kpoints_settings (dict[str, Any]):  The settings
            used to create the k-grid parameters for FHI-aims
        use_structure_charge (bool): If set to True, then the overall charge of the
            structure (structure.charge) is used to set the `charge` variable in the
            `control.in`. Default is False.
    """

    user_params: dict[str, Any] = field(default_factory=dict)
    user_kpoints_settings: dict[str, Any] = field(default_factory=dict)
    use_structure_charge: bool = False

    def get_input_set(
        self,
        structure: Structure | Molecule | None = None,
        prev_dir: PathLike | None = None,
        properties: list[str] | None = None,
    ) -> AimsInputSet:
        """Generate an AimsInputSet object.

        Args:
            structure (Structure or Molecule): Structure or
                Molecule to generate the input set for.
            prev_dir (str or Path): Path to the previous working directory
            properties (list[str]): System properties that are being calculated

        Returns:
            AimsInputSet: The input set for the calculation of structure
        """
        prev_structure, prev_parameters, _ = self._read_previous(prev_dir)

        structure = structure or prev_structure

        if structure is None:
            raise ValueError("No structure can be determined to generate the input set")

        parameters = self._get_input_parameters(structure, prev_parameters)
        properties = self._get_properties(properties, parameters)

        return AimsInputSet(parameters=parameters, structure=structure, properties=properties)

    @staticmethod
    def _read_previous(
        prev_dir: PathLike | None = None,
    ) -> tuple[Structure | Molecule | None, dict[str, Any], dict[str, Any]]:
        """Read in previous results.

        Args:
            prev_dir (str or Path): The previous directory for the calculation
        """
        prev_structure: Structure | Molecule | None = None
        prev_params = {}
        prev_results: dict[str, Any] = {}

        if prev_dir:
            # strip hostname from the directory (not good, works only with run_locally.
            # Should be checked with Fireworks, will not for sure work with
            # jobflow_remote)
            split_prev_dir = str(prev_dir).split(":")[-1]
            with open(f"{split_prev_dir}/parameters.json") as param_file:
                prev_params = json.load(param_file, cls=MontyDecoder)

            try:
                aims_output = AimsOutput.from_outfile(f"{split_prev_dir}/aims.out")
                prev_structure = aims_output.get_results_for_image(-1)

                prev_results = prev_structure.properties
                prev_results.update(prev_structure.site_properties)
            except (IndexError, AimsParseError):
                pass

        return prev_structure, prev_params, prev_results

    @staticmethod
    def _get_properties(
        properties: list[str] | None = None,
        parameters: dict[str, Any] | None = None,
    ) -> list[str]:
        """Get the properties to calculate.

        Args:
            properties (list[str]): The currently requested properties
            parameters (dict[str, Any]): The parameters for this calculation

        Returns:
            list[str]: The list of properties to calculate
        """
        if properties is None:
            properties = ["energy", "free_energy"]

        if parameters is None:
            return properties

        if "compute_forces" in parameters and "forces" not in properties:
            properties.append("forces")
        if "compute_heat_flux" in parameters and "stresses" not in properties:
            properties.append("stress")
            properties.append("stresses")
        if "stress" not in properties and (
            ("compute_analytical_stress" in parameters)
            or ("compute_numerical_stress" in parameters)
            or ("compute_heat_flux" in parameters)
        ):
            properties.append("stress")

        return properties

    def _get_input_parameters(
        self,
        structure: Structure | Molecule,
        prev_parameters: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Create the input parameters.

        Args:
            structure (Structure | Molecule): The structure
                or molecule for the system
            prev_parameters (dict[str, Any]): The previous
                calculation's calculation parameters

        Returns:
            dict: The input object
        """
        # Get the default config
        # FHI-aims recommends using their defaults so bare-bones default params
        params: dict[str, Any] = {"xc": "pbe", "relativistic": "atomic_zora scalar"}

        # Override default parameters with previous parameters
        prev_parameters = {} if prev_parameters is None else copy.deepcopy(prev_parameters)
        prev_parameters.pop("relax_geometry", None)
        prev_parameters.pop("relax_unit_cell", None)

        kpt_settings = copy.deepcopy(self.user_kpoints_settings)
        if isinstance(structure, Structure) and "k_grid" in prev_parameters:
            density = self.k2d(structure, prev_parameters.pop("k_grid"))
            if "density" not in kpt_settings:
                kpt_settings["density"] = density

        parameter_updates = self.get_parameter_updates(structure, prev_parameters)
        params = recursive_update(params, parameter_updates)
        # Add the structure charge (useful for defect workflows)
        if self.use_structure_charge:
            params["charge"] = structure.charge

        # Override default parameters with user_params
        params = recursive_update(params, self.user_params)
        if ("k_grid" in params) and ("density" in kpt_settings):
            warn(
                "WARNING: the k_grid is set in user_params and in the kpt_settings,"
                " using the one passed in user_params.",
                stacklevel=2,
            )
        elif isinstance(structure, Structure) and ("k_grid" not in params):
            density = kpt_settings.get("density", 5.0)
            even = kpt_settings.get("even", True)
            params["k_grid"] = self.d2k(structure, density, even)
        elif isinstance(structure, Molecule) and "k_grid" in params:
            warn("WARNING: removing unnecessary k_grid information", stacklevel=2)
            del params["k_grid"]

        return params

    def get_parameter_updates(
        self,
        structure: Structure | Molecule,
        prev_parameters: dict[str, Any],
    ) -> dict[str, Any]:
        """Update the parameters for a given calculation type.

        Args:
            structure (Structure or Molecule): The system to run
            prev_parameters (dict[str, Any]): Previous calculation parameters.

        Returns:
            dict: A dictionary of updates to apply.
        """
        return prev_parameters

    def d2k(
        self,
        structure: Structure,
        kpt_density: float | tuple[float, float, float] = 5.0,
        even: bool = True,
    ) -> Iterable[float]:
        """Convert k-point density to Monkhorst-Pack grid size.

        inspired by [ase.calculators.calculator.kptdensity2monkhorstpack]

        Args:
            structure (Structure): Contains unit cell and
                information about boundary conditions.
            kpt_density (float | list[float]): Required k-point
                density.  Default value is 5.0 point per Ang^-1.
            even (bool): Round up to even numbers.

        Returns:
            dict: Monkhorst-Pack grid size in all directions
        """
        recip_cell = structure.lattice.inv_matrix.transpose()
        return self.d2k_recip_cell(recip_cell, structure.lattice.pbc, kpt_density, even)

    def k2d(self, structure: Structure, k_grid: np.ndarray[int]):
        """Generate the kpoint density in each direction from given k_grid.

        Args:
            structure: Structure
                Contains unit cell and information about boundary conditions.
            k_grid: np.ndarray[int]
                k_grid that was used.

        Returns:
            dict: Density of kpoints in each direction. result.mean() computes average density
        """
        recip_cell = structure.lattice.inv_matrix.transpose()
        densities = k_grid / (2 * np.pi * np.sqrt((recip_cell**2).sum(axis=1)))
        return np.array(densities)

    @staticmethod
    def d2k_recip_cell(
        recip_cell: np.ndarray,
        pbc: Sequence[bool],
        kpt_density: float | tuple[float, float, float] = 5.0,
        even: bool = True,
    ) -> Sequence[int]:
        """Convert k-point density to Monkhorst-Pack grid size.

        Args:
            recip_cell (Cell): The reciprocal cell
            pbc (Sequence[bool]): If element of pbc is True
                then system is periodic in that direction
            kpt_density (float or list[floats]): Required k-point
                density. Default value is 5 points per Ang^-1.
        even(bool): Round up to even numbers.

        Returns:
            dict: Monkhorst-Pack grid size in all directions
        """
        if isinstance(kpt_density, float):
            kpt_density = (kpt_density, kpt_density, kpt_density)
        kpts: list[int] = []
        for i in range(3):
            if pbc[i]:
                k = 2 * np.pi * np.sqrt((recip_cell[i] ** 2).sum()) * float(kpt_density[i])
                if even:
                    kpts.append(2 * int(np.ceil(k / 2)))
                else:
                    kpts.append(int(np.ceil(k)))
            else:
                kpts.append(1)
        return kpts


def recursive_update(dct: dict, up: dict) -> dict:
    """
    Update a dictionary recursively and return it.

    Args:
        dct (dict): Input dictionary to modify
        up (dict): updates to apply

    Returns:
        dict: The updated dictionary.

    Example:
        d = {'activate_hybrid': {"hybrid_functional": "HSE06"}}
        u = {'activate_hybrid': {"cutoff_radius": 8}}

        yields {'activate_hybrid': {"hybrid_functional": "HSE06", "cutoff_radius": 8}}}
    """
    for key, val in up.items():
        if isinstance(val, dict):
            dct[key] = recursive_update(dct.get(key, {}), val)
        elif key == "output" and isinstance(val, list):  # for all other keys the list addition is not needed (I guess)
            old_v = dct.get(key, [])
            dct[key] = old_v + val
        else:
            dct[key] = val
    return dct



"""An object representing parse results of FHI-aims standard output."""

import os.path
import warnings
from dataclasses import dataclass, field
from pathlib import Path
from typing import IO, Any

from pyfhiaims.errors import AimsParseError
from pyfhiaims.geometry import AimsGeometry
from pyfhiaims.outputs.parser import StdoutParser
from pyfhiaims.outputs.parser.utils import NamedDict

CONVERGED_RESULTS_KEYS = [
    "total_energy",
    "free_energy",
    "forces",
    "stress",
    "stresses",
    "magmom",
    "dipole",
    "fermi_energy",
    "n_scf_iter",
    "mulliken_charges",
    "mulliken_spins",
    "hirshfeld_charges",
    "hirshfeld_dipole",
    "hirshfeld_volumes",
    "hirshfeld_atomic_dipoles",
    "dielectric_tensor",
    "polarization",
    "vbm",
    "cbm",
    "gap",
    "direct_gap",
    "pressure",
]


@dataclass
class AimsImage:
    """Object to represent the aims outputs."""

    _geometry: AimsGeometry = None
    _results: dict[str, Any] = field(default=dict)
    _scf: list[dict[str, Any]] = field(default=list)
    _converged: bool = None

    @classmethod
    def from_spec(cls, ionic_spec: dict[str, Any]):
        """Get the image from the image spec."""
        scf = ionic_spec.pop("scf_steps", [])
        if scf:
            spec = {k: v for k, v in scf[-1].items() if "change" not in k}
        else:
            warnings.warn(UserWarning("No SCF steps found"), stacklevel=2)
            spec = {}

        spec.update(ionic_spec)

        geometry = spec.pop("geometry", None)
        converged = spec.pop("scf_converged", False)

        results: dict[str, Any] = {
            key.replace("atomic_", ""): spec.get(key) for key in spec
        }

        if (geometry and geometry.lattice_vectors is not None
                and spec.get("is_metallic", False)):
            spec["total_energy"] = results["total_energy"]
            results["total_energy"] = spec.get("corrected_total_energy")

        results["n_scf_iter"] = len(scf)
        results["magmom"] = spec.get("spin")
        results["fermi_energy"] = spec.get("chemical_potential")
        results["stress"] = spec.get(
            "analytical_stress", spec.get("numerical_stress")
        )
        results["dipole"] = spec.get("total_dipole_moment")
        results["absolute_dipole"] = spec.get("absolute_dipole_moment")
        results["energy"] = results["free_energy"]

        return cls(
            _geometry=geometry,
            _results=results,
            _scf=scf,
            _converged=converged,
        )

    def __post_init__(self):
        """Check if lattice exists, and if so make fractional coords."""
        if self._geometry and self._geometry.lattice_vectors is not None:
            for atom in self._geometry.atoms:
                atom.set_fractional(self._geometry.lattice_vectors)

    def __getitem__(self, key: str) -> Any:
        """Get a particular result."""
        try:
            return self._results[key]
        except KeyError:
            if key in CONVERGED_RESULTS_KEYS:
                return None
            raise ValueError(f"Requested item ({key}) is not a parsable result")

    def __getattr__(self, name: str) -> Any:
        """Get a particular result for the result."""
        return self[name]

    @property
    def geometry(self):
        """Return the geometry of the calculation."""
        return self._geometry

    @property
    def converged(self):
        """Return True if calculation is converged."""
        return self._converged

    def get_results(self, verbosity: str = "converged") -> dict[str, Any]:
        """Return the results dictionary."""
        if verbosity not in ["converged", "all"]:
            raise ValueError(
                f"verbosity must be one of 'converged' or 'all', not {verbosity}"
            )
        if verbosity == "converged":
            return {k: v for k, v in self._results.items()
                    if k in CONVERGED_RESULTS_KEYS}
        self._results["scf_steps"] = self._scf
        return self._results

    @property
    def results(self):
        """Return the results dictionary."""
        return self.get_results(verbosity="converged")

    @property
    def scf(self):
        """Return the SCF data."""
        return self._scf


class AimsStdout:
    """The standard output parser and results handler for FHI-aims calculations.

    This class is responsible for parsing the output file of an FHI-aims calculation
    and storing the extracted results in a structured format. It facilitates access
    to parsed results, metadata, warnings, and errors associated with the calculation.

    Attributes:
        file_name (str): The full path to the FHI-aims output file being processed.

    Methods:
        results (dict[str, Any]): Provides parsed results for FHI-aims calculation.
        metadata (NamedDict[str, Any]): Provides runtime choices and background
            calculation metadata.
        warnings (list[str]): Retrieves warning messages found in the output file.
        errors (list[str]): Retrieves error messages from the output, typically at the
            end of the file.
        is_finished_ok (bool): Checks if the calculation finished successfully.
        __getattr__: Allows accessing results dictionary
            attributes as object attributes.
        keys (set[str]): Retrieves the set of all keys present in the output object.

    Raises:
        FileNotFoundError: If the specified output file does not exist.
        AimsParseError: If the output file is not a valid FHI-aims output.

    """

    def __init__(
        self,
        stdout_file: str | Path | IO,
        parser: StdoutParser = None,
    ):
        """Construct AimsStdout.

        Args:
            stdout_file (str | Path | IO): The output file
            parser (StdoutParser, optional): The parser to use.

        """
        if isinstance(stdout_file, str | Path):
            self.file_name = Path(stdout_file).resolve().as_posix()
            file_path = Path(stdout_file)
            if not file_path.is_file():
                raise FileNotFoundError(
                    f"FHI-aims output file {self.file_name} does not exist"
                )
        else:
            self.file_name = os.path.realpath(stdout_file.name)
        self._parser = parser or StdoutParser(stdout_file)
        self._results = self._parser.parse()

        # gather the parsed results into one place, like in ASE parser by Tom
        if "aims_version" not in self._results:
            raise AimsParseError(
                f"{self.file_name} is not a valid FHI-aims output file"
            )

        if len(self._results.get("ionic_steps", [])) > 0:
            ionic_steps = self._results["ionic_steps"]
        else:
            warnings.warn(
                UserWarning("FHI-aims calculation did not start"), stacklevel=2
            )
            ionic_steps = []

        specs = [spec.copy() for spec in ionic_steps]

        if len(specs) > 0:
            if self.metadata.get("relax") or self.metadata.get("calculate_atom_bsse"):
                for ii in range(1, len(specs)):
                    specs[ii]["geometry"] = specs[ii - 1]["geometry"]

            init_geo = AimsGeometry.from_strings(
                self._results["input"]["geometry_in"].splitlines()
            )
            if init_geo.lattice_vectors is not None:
                for atom in init_geo.atoms:
                    atom.set_fractional(init_geo.lattice_vectors)
            specs[0]["geometry"] = init_geo

        if "final" not in self._results:
            self._results["final"] = {}

        if "time" in self._results:
            self._results["final"]["cpu_time"] = self._results["time"]["total"][0]
            self._results["final"]["total_time"] = self._results["time"]["total"][1]

        self._results["final"]["dielectric_tensor"] = None
        if "dfpt" in self._results:
            self._results["final"]["dielectric_tensor"] = self._results["dfpt"][
                "dielectric_tensor"
            ]
        if len(specs) > 0:
            specs[-1].update(self.final)
        # this is not a good style, as AimsStdout should not depend on Image
        # also, this is just erroneous, as there are a lot of situations
        # when the result is not trajectory-like, e.g. numerical stress calculation
        self._images = [AimsImage.from_spec(spec) for spec in specs]
        if len(self._images) > 0:
            self._results["final"].update(self._images[-1].results)

    @property
    def results(self) -> dict[str, Any]:
        """A dictionary with results for FHI-aims calculation."""
        return self._results

    @property
    def images(self) -> list[AimsImage]:
        """Get all images in the file."""
        return self._images

    @property
    def n_images(self) -> int:
        """Get all images in the file."""
        return len(self._images)

    @property
    def metadata_summary(self) -> dict[str, Any]:
        """Summarize metadata into a single object."""
        return {
            "commit_hash": self._results.get("commit_hash"),
            "aims_uuid": self._results.get("aims_uuid"),
            "version_number": self._results.get("aims_version"),
            "fortran_compiler": self._results.get("fortran_compiler", "").split("/")[
                -1
            ],
            "c_compiler": self._results.get("c_compiler", "").split("/")[-1],
            "fortran_compiler_flags": self._results.get("fortran_compiler_flags"),
            "c_compiler_flags": self._results.get("c_compiler_flags"),
            "build_type": self._results.get("build_type"),
            "linked_against": self._results.get("linked_against"),
        }

    @property
    def header_summary(self) -> dict[str, Any]:
        """Summarize the initial structure into a single object."""
        init_geo = AimsGeometry.from_strings(
            self._results["input"]["geometry_in"].splitlines()
        )
        return {
            "initial_geometry": init_geo,
            "initial_lattice": init_geo.lattice_vectors,
            "is_relaxation": self.metadata.relax,
            "is_md": self.metadata.md,
            "n_atoms": self.metadata.num_atoms,
            "n_bands": self.metadata.num_bands,
            "n_electrons": int(self.metadata.num_electrons),
            "n_spins": self.metadata.num_spins,
            "electronic_temperature": self._results.get("electronic_temperature", 0.0),
            "n_k_points": self.metadata.get("num_k_points"),
            "k_points": self._results.get("k_points"),
            "k_point_weights": self._results.get("k_point_weights"),
        }

    def get_image(self, ind: int) -> AimsImage:
        """Get the image at index ind."""
        return self._images[ind]

    @property
    def metadata(self) -> NamedDict[str, Any]:
        """A metadata dictionary for FHI-aims calculation,
        including runtime choices and some background calculation checks.
        """
        return NamedDict(self._parser.run_metadata)

    @property
    def warnings(self) -> list[str]:
        """A list of warning messages for FHI-aims calculation.

        A warning message is the one beginning with an optional space and one or
        several asterisks.
        """
        return self._parser.warnings

    @property
    def errors(self) -> list[str]:
        """A list of error messages for FHI-aims calculation.

        An error message is a warning at the end of the file.
        """
        return self._parser.errors

    @property
    def is_finished_ok(self) -> bool:
        """A check if the calculation is finished successfully."""
        return self._results.get("is_finished_ok", False)

    def __getitem__(self, ind: int) -> AimsImage:
        """Get an image."""
        return self.get_image(ind)

    def __len__(self) -> int:
        """Get n_images."""
        return self.n_images

    def __getattr__(self, item) -> Any:
        """Get a value for a specific item."""
        if item in self._results["final"]:
            return self._results["final"][item]
        if item in self._results:
            return self._results[item]
        return None

    def keys(self) -> set[str]:
        """Return the keys present in the output object."""
        return set(
            list(self._results.get("final", {}).keys())
            + list(self._results.keys())
            + ["metadata", "warnings", "errors"]
        )



"""Definiton of the BasisSet objects for SpeciesDefinitions."""

from __future__ import annotations

from dataclasses import dataclass
from warnings import warn

from pyfhiaims.species_defaults.basis_function import BASIS_TYPES, BasisFunction
from pyfhiaims.utils import is_number_string


class TierNotFoundError(Exception):
    """Exception raised if requested tier does not exist."""


_TIERS = {
    "First tier": 1,
    "Second tier": 2,
    "Third tier": 3,
    "Fourth tier": 4,
    "Further basis functions": 5,
    "User added via pyaims.species_defaults.BasisSet": 6,
}


@dataclass
class BasisSet:
    """A basis set class for an FHI-aims species' defaults.

    Parameters
    ----------
    basis_set: list[list[BasisFunction]]

    """

    basis_set: list[list[BasisFunction]]

    def __str__(self):
        """Get the string for the basis set."""
        return self.to_string()

    @classmethod
    def from_string(cls, string: str) -> BasisSet:
        """Create a basis set instance from text block."""
        basis_set = [[]]
        lines = string.splitlines()
        tier = 1
        is_gaussian = False

        basis_lines = []
        for line in lines:
            line = line.strip()
            # check for the empty string
            if not line:
                continue
            # check for tier increase
            if line.startswith("#"):
                for k, tier_i in _TIERS.items():
                    if k in line:
                        tier = tier_i
                        if len(basis_set) < tier:
                            basis_set += [[] for _ in range(tier - len(basis_set))]
                        break
                # check if the commented line contains the disabled basis function
                words = line.replace("#", "").split()
                if len(words) < 1 or (
                    not any(bt in words[0] for bt in BASIS_TYPES)
                    and not is_number_string(line)
                ):
                    continue
            # gaussian is the only basis function that can span several lines
            if not is_gaussian:
                basis_lines.append(line)

            if not is_gaussian and "gaussian" in line:
                is_gaussian = True
                continue

            if is_gaussian and is_number_string(line):
                basis_lines.append(line)
                continue
            if is_gaussian:
                basis_set[tier - 1].append(
                    BasisFunction.from_string("\n".join(basis_lines))
                )
                is_gaussian = "gaussian" in line
                basis_lines = [line]

            if is_gaussian:
                continue
            basis_set[tier - 1].append(
                BasisFunction.from_string("\n".join(basis_lines))
            )
            basis_lines = []
        return cls(basis_set=basis_set)

    @property
    def n_tiers(self) -> int:
        """Returns the number of tiers in the set. Noise level functions always
        belong to Tier 5.
        """
        return len(self.basis_set)

    def tier(
        self, n: int, enabled: bool | None = None, aux: bool = False
    ) -> list[BasisFunction]:
        """Return all basis functions pertaining to the current tier.

        Parameters
        ----------
        n: int
            Ordinal number of the tier. First -> 1, Second -> 2, etc.
        enabled: bool | None
            A flag indicating whether only enabled or only disabled functions
            should be returned. If None, all functions are returned.
        aux: bool
            A flag indicating that aux functions should be returned.

        """
        if n < 1 or n > self.n_tiers:
            raise TierNotFoundError(
                f"No tier with ordinal value {n}. Tier value is [1, {self.n_tiers}]."
            )
        basis_tier = self.basis_set[n - 1]
        if enabled is None:
            return [f for f in basis_tier if f.aux == aux]

        return [f for f in basis_tier if (f.enabled == enabled and f.aux == aux)]

    def _switch_tier(
        self,
        tier: int,
        n: int | str = "all",
        aux: bool = False,
        enable: bool = True,
    ) -> None:
        """Swith the basis functions in the tier to on or off state.

        Parameters
        ----------
        tier: int
            Ordinal number of the tier you want to activate,
            First -> 1, Second -> 2, etc.
        n: int | str
            The number of functions you want to switch in that tier (`all` for entire)
        aux: bool
            the flag indicating if auxiliary functions should be switched.
        enable: bool
            A flag indicating in which state should the basis functions be.

        """
        if tier < 1 or tier > self.n_tiers:
            raise TierNotFoundError(
                f"No tier with ordinal value {tier}. Tier value is [1, {self.n_tiers}]."
            )

        if isinstance(n, int) and n > len(self.tier(tier, enabled=not enable, aux=aux)):
            warn(
                f"Requesting to {'de' if not enable else ''}activate "
                f"more functions in a tier than possible, just doing all.",
                RuntimeWarning,
                stacklevel=2,
            )
            n = "all"

        n = len(self.tier(tier, enabled=not enable, aux=aux)) if n == "all" else n
        i = 0
        for basis_function in self.basis_set[tier - 1][:: (-1) ** (not enable)]:
            if basis_function.enabled == (not enable) and basis_function.aux == aux:
                basis_function.enabled = enable
                i += 1
            if i == n:
                return

    def activate_tier(self, tier: int, n: int | str = "all", aux: bool = False) -> None:
        """Move basis functions in a specified tier from inactive to active.

        Parameters
        ----------
        tier: int
            Ordinal number of the tier you want to activate,
            First -> 1, Second -> 2, etc.
        n: int | str
            The number of functions you want to activate in that tier
            (all for entire)
        aux: bool
            the flag indicating if auxiliary functions should be activated.

        """
        self._switch_tier(tier, n, aux, enable=True)

    def deactivate_tier(
        self, tier: int, n: int | str = "all", aux: bool = False
    ) -> None:
        """Move basis functions in a specified tier from active to inactive.

        Parameters
        ----------
        tier: int
            Ordinal number of the tier you want to deactivate,
            First -> 1, Second -> 2, etc.
        n: int | str
            The number of functions you want to deactivate in that tier
            (all for entire)
        aux: bool
            the flag indicating if auxiliary functions should be activated.

        """
        self._switch_tier(tier, n, aux, enable=False)

    def _switch_functions(
        self, n: int | str = "all", aux: bool = False, enable: bool = True
    ) -> None:
        """Swith a basis functions (disregarding the tier) on or off.

        Parameters
        ----------
        n: int | str
            The number of functions you want to activate (all for entire basis_set)
        aux: bool
            the flag indicating if auxiliary functions should be switched.
        enable: bool
            A flag indicating in which state should the basis functions be.

        """
        valid_functions = sum(
            [len(self.tier(i + 1, not enable, aux)) for i in range(self.n_tiers)]
        )
        if isinstance(n, int) and n > valid_functions:
            warn(
                f"Requesting to {'de' if not enable else ''}activate "
                f"more functions than possible, just doing all.",
                RuntimeWarning,
                stacklevel=2,
            )
            n = "all"

        n = valid_functions if n == "all" else n
        if not enable:
            tier_list = list(range(self.n_tiers, 0, -1))
        else:
            tier_list = list(range(1, self.n_tiers + 1))
        for tier in tier_list:
            valid_fs_tier = len(self.tier(tier, not enable, aux))
            self._switch_tier(tier, min(n, valid_fs_tier), aux, enable)
            if valid_fs_tier < n:
                n -= valid_fs_tier
            else:
                return

    def activate_functions(self, n: int | str = "all", aux: bool = False) -> None:
        """Move basis functions from inactive to active.

        Parameters
        ----------
        n: int | str
            The number of functions you want to activate (all for entire basis_set)
        aux: bool
            the flag indicating if auxiliary functions should be activated.

        """
        self._switch_functions(n, aux, enable=True)

    def deactivate_functions(self, n: int | str = "all", aux: bool = False) -> None:
        """Move basis functions from active to inactive.

        Parameters
        ----------
        n: int | str
            The number of functions you want to activate (all for entire basis_set)
        aux: bool
            the flag indicating if auxiliary functions should be activated.

        """
        self._switch_functions(n, aux, enable=False)

    def add_function(
        self,
        basis_function: BasisFunction,
        tier: int | None = None,
        active: bool = True,
        aux: bool = False,
    ):
        """Add a basis function to a tier.

        Parameters
        ----------
        basis_function: BasisFunction
            The basis function to add
        tier: int
            The tier to add it to (if None, add to the last `User added` tier)
        active: bool
            If True then add the function to the active list
        aux: bool
            If True then add the function to the auxiliary basis set

        """
        if tier is not None and (tier < 1 or tier > self.n_tiers):
            raise TierNotFoundError(
                f"No tier with ordinal value {tier}. Tier value is [1, {self.n_tiers}]."
            )
        basis_function.enabled = active
        basis_function.aux = aux

        if tier is not None:
            self.basis_set[tier - 1].append(basis_function)
        else:
            user_tier = _TIERS["User added via pyaims.species_defaults.BasisSet"]
            if self.n_tiers < user_tier:
                self.basis_set += [[] for _ in range(user_tier - self.n_tiers)]
                self.basis_set[user_tier - 1].append(basis_function)

    def to_string(self):
        """Get the text block for the basis set."""
        comments_to_tiers = {v: k for k, v in _TIERS.items()}
        result = []
        for tier in range(1, self.n_tiers + 1):
            if self.basis_set[tier - 1]:
                result.append(f"#  {comments_to_tiers[tier]}")
                for basis_function in self.basis_set[tier - 1]:
                    result.append(str(basis_function))
        return "\n".join(result)



"""Different exceptions pertaining to pyaims."""


class PyaimsError(Exception):
    """A general Pyaims exception."""


class InvalidSpeciesInput(PyaimsError):
    """Exception raised if an error occurs when creating species' defaults."""


class AimsParseError(PyaimsError):
    """Exception raised if an error occurs when parsing an Aims output file."""



"""Test the AimsCube interface"""

import numpy as np
import pytest

from pyfhiaims.control.cube import (
    ALLOWED_AIMS_CUBE_FORMATS,
    ALLOWED_AIMS_CUBE_TYPES,
    ALLOWED_AIMS_CUBE_TYPES_STATE,
    AimsCube,
)

content_block = """output cube elf
    cube origin -1.000000000000e-01  2.000000000000e-01  3.000000000000e-01
    cube edge 30  2.000000000000e-01  1.000000000000e-01  5.000000000000e-02
    cube edge 40 -2.000000000000e-01  1.000000000000e-01 -1.000000000000e-02
    cube edge 10  3.000000000000e-01  2.000000000000e-01  1.000000000000e-01
    cube format cube
    cube spinstate 1
    cube kpoint 2
    cube filename cube_file.cub
    cube elf_type 0
"""


def test_cube(tmp_path):
    """Test the AimsCube Interface"""
    cube_1 = AimsCube(
        type="elf",
        origin=[-0.1, 0.2, 0.3],
        edges=[[0.2, 0.1, 0.05], [-0.2, 0.1, -0.01], [0.3, 0.2, 0.1]],
        points=[30, 40, 10],
        spin_state=1,
        kpoint=2,
        filename="cube_file.cub",
        format="cube",
        elf_type=0,
    )

    test_neq = {
        "type": "stm",
        "origin": [0.1, 0.2, 0.3],
        "edges": [[0.1, 0.1, 0.05], [-0.2, 0.1, -0.01], [0.3, 0.2, 0.1]],
        "points": [20, 40, 10],
        "spin_state": 2,
        "kpoint": 1,
        "filename": "test_cube_file.cub",
        "format": "xsf",
        "elf_type": 1,
    }

    assert cube_1.control_block == content_block
    cube_2 = AimsCube.from_dict(cube_1.as_dict())

    assert cube_1 == cube_2
    for key, val in test_neq.items():
        setattr(cube_2, key, val)
        assert cube_1 != cube_2

        setattr(cube_2, key, getattr(cube_1, key))
        assert cube_1 == cube_2

    for typ in ALLOWED_AIMS_CUBE_TYPES:
        _ = AimsCube(type=typ)
        with pytest.raises(
            ValueError,
            match="can not have a state associated with it"
        ):
            _ = AimsCube(type=f"{typ} 0")

        if typ != "elf":
            with pytest.raises(
                ValueError,
                match="elf_type is only used when the cube type is elf"
            ):
                _ = AimsCube(type=typ, elf_type=1)

    for typ in ALLOWED_AIMS_CUBE_TYPES_STATE:
        _ = AimsCube(type=f"{typ} 1")

        with pytest.raises(
            ValueError,
            match="must have a state associated with it"
        ):
            _ = AimsCube(type=f"{typ}")

        with pytest.raises(
            ValueError,
            match="must have a state associated with it"
        ):
            _ = AimsCube(type=f"{typ} 1 1")

        with pytest.raises(
            ValueError,
            match="elf_type is only used when the cube type is elf"
        ):
            _ = AimsCube(type=f"{typ} 1", elf_type=1)

    with pytest.raises(
        ValueError,
        match="Cube type undefined"
    ):
        _ = AimsCube(type="asdfg")

    for cb_format in ALLOWED_AIMS_CUBE_FORMATS:
        _ = AimsCube(type="total_density", format=cb_format)

    with pytest.raises(
        ValueError,
        match="is invalid. Cube files must have a format of"
    ):
        _ = AimsCube(type="total_density", format="asdfg")

    with pytest.raises(
        ValueError,
        match="Spin state must be one of"
    ):
        _ = AimsCube(type="total_density", spin_state=3)

    with pytest.raises(
        ValueError,
        match="he cube origin must have 3 components"
    ):
        _ = AimsCube(type="total_density", origin=[0])

    with pytest.raises(
        ValueError,
        match="The number of points per edge must have 3 components"
    ):
        _ = AimsCube(type="total_density", points=[0])

    with pytest.raises(
        ValueError,
        match="Only three cube edges can be passed"
    ):
        _ = AimsCube(type="total_density", edges=np.eye(4) * 0.1)

    edges = test_neq["edges"]
    edges[2].append(3)
    with pytest.raises(
        ValueError,
        match="Each cube edge must have 3 components"
    ):
        _ = AimsCube(type="total_density", edges=edges)



"""Define the atom object for FHI-aims geometry files."""

from __future__ import annotations

import re
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

import numpy as np

from .periodic_table import ATOMIC_SYMBOLS_TO_NUMBERS

if TYPE_CHECKING:
    from pyfhiaims.utils.typecast import Matrix3D, Vector3D


class FHIAimsAtomType(Enum):
    """Enum type for atoms."""

    ATOM = 1
    EMPTY = 2
    PSEUDOCORE = 3


@dataclass
class FHIAimsAtom:
    """Atom object for FHI-aims.

    Parameters
    ----------
    symbol: Sequence[str]
        List of symbols for each nucleus in the structure
    number: Optional[Sequence[int]]
        Atomic number for the nuclei
    position: Sequence[Vector3D]
        The position of each nucleus in space
    fractional_position: Optional[Sequence[Vector3D]]
        The fractional_position of the atoms
    velocity: Optional[Sequence[Vector3D]]
        The velocity of each nucleus in space
    initial_charge: Sequence[float]
        The initial charge for each nuclei
    initial_moment: Sequence[float]
        The initial magnetic moment for each nuclei
    constraints: Sequence[tuple[bool, bool, bool]]
        Fix_atom constraints for each nucleus and direction
    is_empty: Optional[Sequence[bool]]
        Specifies if site is for an empty_site
    constraint_region: Optional[Sequence[int | None]]
        Assigns the immediately preceding atom to the region labelled number.
        number is the integer number of a spatial region, which must correspond to a
        region defined by keyword constraint_electrons in file control.in
    is_pseudocore: Optional[Sequence[bool]]
        True if site is a pseudocore
    magnetic_response: Optional[Sequence[bool]]
        Includes the current atom in the magnetic response calculations. If only
        the magnetizability is required, this keyword need not be used in geometry.in.
        Otherwise, the calculation of the shieldings or J-couplings is aborted if no
        atoms are flagged for MR calculations in geometry.in
    magnetic_moment: Optional[Sequence[float]]
        Overrides the default magnetic moment for the given atom. The default values
        (in units of the nuclear magneton µN ) can be found in
        MagneticResponse/MR_nuclear_data.f90. In case of J-couplings, the isotopes
        used are also printed in the output.
    nuclear_spin: Optional[Sequence[float]]
        Overrides the default nuclear spin for the given atom. The default
        values can be found in MagneticResponse/MR_nuclear_data.f90 and are also
        printed in the output for J-coupling calculations.
    isotope: Optional[Sequence[int]]
        Overrides the default isotope mass number for the given atom. For
        more flexibility, the magnetic moment and spin can be specified separately
        with the above keywords. The default isotopes numbers can be found in
        MagneticResponse/MR_nuclear_data.f90.
    RT_TDDFT_initial_velocity: Sequence[Vector3D]
        Initial velocity of corresponding (i.e. last specified) atom when
        peforming RT-TDDFT-Ehrenfest dynamics

    """

    symbol: str = None
    number: int | None = None
    position: Vector3D = None
    fractional_position: Vector3D | None = None
    velocity: Vector3D | None = None
    initial_charge: float = 0.0
    initial_moment: float = 0.0
    constraints: tuple[bool, bool, bool] | None = None
    constraint_region: int | None = None
    magnetic_response: bool | None = None
    magnetic_moment: float | None = None
    nuclear_spin: float | None = None
    isotope: int | None = None
    is_empty: bool = None
    is_pseudocore: bool = None
    RT_TDDFT_initial_velocity: Vector3D | None = None

    def __post_init__(self):
        symbol = re.search(r"[A-Za-z]+", self.symbol).group(0)
        self.number = ATOMIC_SYMBOLS_TO_NUMBERS[symbol]

    def set_fractional(self, lattice_vectors: Matrix3D, wrap: bool = False):
        """Set the fractional postion given lattice vectors.

        Parameters
        ----------
        lattice_vectors: Matrix3D
            The lattice vectors for ths system
        wrap: bool
            True it to wrap atoms into the unit cell

        """
        self.fractional_position = np.linalg.solve(
            lattice_vectors.T, np.transpose(self.position)
        ).T
        if wrap:
            self.fractional_position %= 1.0
            self.fractional_position %= 1.0

    def to_string(self):
        """Convert an FHIAimsAtom object to geometry.in string."""
        pos = self.position
        # what happens if both empty and pseudocore are chosen?
        if self.is_empty:
            atom_str = "empty     "
        elif self.is_pseudocore:
            atom_str = "pseudocore"
        elif self.fractional_position is not None:
            atom_str = "atom_frac "
            pos = self.fractional_position
        else:
            atom_str = "atom      "

        content_str = [
            f"{atom_str} "
            + " ".join([f"{pi:>20.12e}" for pi in pos])
            + f" {self.symbol}"
        ]

        if self.velocity is not None:
            content_str.append(
                "    velocity " + " ".join([f"{vel:>20.12e}" for vel in self.velocity])
            )

        if np.abs(self.initial_charge) > 1e-12:
            content_str.append(f"    initial_charge {self.initial_charge:.12f}")

        if self.initial_moment > 1e-12:
            content_str.append(f"    initial_moment {self.initial_moment:.12f}")

        if self.constraints is not None and np.any(self.constraints):
            content_str.append(
                str(
                    f"    constrain_relaxation "
                    f"{'x ' if self.constraints[0] else ''}"
                    f"{'y ' if self.constraints[1] else ''}"
                    f"{'z' if self.constraints[2] else ''}"
                ).rstrip()
            )

        if self.constraint_region is not None and self.constraint_region > 0:
            content_str.append(f"    constraint_region {self.constraint_region}")

        if self.magnetic_response is not None and self.magnetic_response:
            content_str.append("    magnetic_response")
        if self.magnetic_moment is not None and np.abs(self.magnetic_moment) > 1e-12:
            content_str.append(f"    magnetic_moment {self.magnetic_moment:.12f}")
        if self.nuclear_spin is not None and np.abs(self.nuclear_spin) > 1e-12:
            content_str.append(f"    nuclear_spin {self.nuclear_spin:.12f}")
        if self.isotope is not None and self.isotope > 0:
            content_str.append(f"    isotope {self.isotope}")
        if self.RT_TDDFT_initial_velocity is not None:
            content_str.append(
                "    RT_TDDFT_initial_velocity "
                + " ".join([f"{vel:>20.12e}" for vel in self.RT_TDDFT_initial_velocity])
            )

        return "\n".join(content_str)



"""Tests for basis_function module."""

import pytest

from pyfhiaims.species_defaults.basis_function import (
    BasisFunction,
    BasisFunctionError,
    ConfinedBasisFunction,
    GaussianBasisFunction,
    HydroBasisFunction,
    IonicBasisFunction,
    STOBasisFunction,
)


def test_basis_function():
    """Tests a generic BasisFunction constructor."""
    string = "    hydro 4 f 5.6"
    basis_function = BasisFunction.from_string(string)
    assert basis_function.type == "hydro"
    assert basis_function.enabled
    assert not basis_function.aux

    string = "#    hydro 4 f 5.6"
    basis_function = BasisFunction.from_string(string)
    assert basis_function.type == "hydro"
    assert not basis_function.enabled
    assert not basis_function.aux

    string = "# for_aux hydro 4 f 5.6"
    basis_function = BasisFunction.from_string(string)
    assert basis_function.type == "hydro"
    assert not basis_function.enabled
    assert basis_function.aux

    with pytest.raises(
        ValueError,
        match="No valid basis function type found in"
    ):
        _ = BasisFunction.from_string("asdfg 0 1")


def test_hydro_basis_function():
    """Tests hydrogen-like BasisFunction constructor."""
    string = "    hydro 4 f 5.6"
    basis_function = HydroBasisFunction.from_string(string)
    assert basis_function.type == "hydro"
    assert basis_function.z_eff == 5.6


def test_ionic_basis_function():
    """Tests ionic BasisFunction constructor."""
    string = "    ionic 4 d auto"
    basis_function = IonicBasisFunction.from_string(string)
    assert basis_function.n == 4
    assert basis_function.radius == "auto"
    string = "    ionic 3 d 2.5"
    basis_function = IonicBasisFunction.from_string(string)
    assert basis_function.n == 3
    assert basis_function.radius == 2.5


def test_confined_basis():
    """Test the confined basis function"""
    string = "confined      3  s  0.99"
    basis_function = ConfinedBasisFunction.from_string(string)
    assert basis_function.enabled
    assert basis_function.type == "confined"
    assert basis_function.n == 3
    assert basis_function.l == "s"
    assert basis_function.radius == 0.99
    assert str(basis_function).strip() == string

    string = "# confined 3 s 0.99"
    basis_function = ConfinedBasisFunction.from_string(string)
    assert not basis_function.enabled


gaussian_str_long = """gaussian    1       7
    315.9000000  0.0039266
    74.4200000  0.0298811
    23.4800000  0.1272120
    8.4880000  0.3209430
    3.2170000  0.4554290
    1.2290000  0.2685630
    0.2964000  0.0188336"""


def test_gaussian_basis_function():
    """Tests Gaussian-based BasisFunction constructor."""
    string = "gaussian    0       1  10555500.0000000"
    basis_function = GaussianBasisFunction.from_string(string)
    assert basis_function.enabled
    assert basis_function.type == "gaussian"
    assert basis_function.n == 1
    assert len(basis_function.alpha_i) == 1
    assert not basis_function.aux
    assert basis_function.alpha_i[0] == 1.05555e7
    assert str(basis_function).strip() == string

    string = "# gaussian 0 1  0.105555E+08"
    basis_function = BasisFunction.from_string(string)
    assert not basis_function.enabled

    basis_function = GaussianBasisFunction.from_string(gaussian_str_long)
    assert str(basis_function) == gaussian_str_long

    string = "aux_gaussian 0 1 0.10555E+08"
    basis_function = BasisFunction.from_string(string)
    assert basis_function.aux

    with pytest.raises(BasisFunctionError):
        _ = GaussianBasisFunction(n=2, l=0, alpha_i=[0, 1], coeff_i=None)

    with pytest.raises(BasisFunctionError):
        _ = GaussianBasisFunction(n=2, l=0, alpha_i=[0, 1], coeff_i=[0, 1, 2])

    with pytest.raises(BasisFunctionError):
        _ = GaussianBasisFunction(n=2, l=0, alpha_i=[0, 1, 2], coeff_i=[0, 1])


def test_sto_basis_function():
    """Test the STO basis function"""
    string = "sto      1  0  0.76"
    basis_function = STOBasisFunction.from_string(string)
    assert basis_function.enabled
    assert basis_function.type == "sto"
    assert basis_function.n == 1
    assert basis_function.l == 0
    assert basis_function.zeta == 0.76
    assert str(basis_function).strip() == string

    string = "# sto 1 0   0.76"
    basis_function = STOBasisFunction.from_string(string)
    assert not basis_function.enabled



"""Test new FHI-aims standard output parser."""

from io import StringIO
from textwrap import dedent

import yaml

from pyfhiaims.outputs.parser import StdoutParser
from tests.simple_parser import SimpleParser
from tests.utils import is_subset


def test_empty_parser():
    # empty string - nothing is parsed
    parser = SimpleParser(StringIO(""))
    result = parser.parse()
    assert result == {}
    # Not a parsable string - nothing is parsed
    parser = SimpleParser(StringIO("asdfghjkl"))
    result = parser.parse()
    assert result == {}


def test_converters():
    file_str = dedent("""\
    Version: 1.2
    --- B CHUNK ---
     | B1 = 1
     | B2 = 2 Ry 3.5 eV
     some useless lines
     B3 = 3.5E-5
    """)
    parser = SimpleParser(StringIO(file_str))
    result = parser.parse()
    assert result["version"] == "1.2"
    assert result["b_values"]["b1"] == 1
    assert result["b_values"]["b2"] == 3.5
    assert result["b_values"]["b3"] == 0.000035
    assert "b4" not in result["b_values"]


def test_results_structure():
    file_str = dedent("""\
    B CHUNK
    some useless lines
     B3 = 3.5E-5
    more useless lines
     | B1 = 2
    --- C 1 ---
    time = 0.01
    --- C 2 ---
    time = 0.02
    Everything is good
    --- Values Converged ---
    End time: 170305.22
    """)
    parser = SimpleParser(StringIO(file_str))
    result = parser.parse()
    assert "version" not in result
    assert result["b_values"]["b1"] == 2
    assert result["b_values"]["b3"] == 0.000035
    assert "b4" not in result["b_values"]
    assert [x["t"] for x in result["b_values"]["c_values"]] == [0.01, 0.02]
    assert [x["all_good"] for x in result["b_values"]["c_values"]] == [False, True]
    assert result["b_values"]["c_values"][-1]["end_time"].second == 5


def test_results_k_points(data_dir):
    """Test parsing of output files."""
    file_name = data_dir / "stdout" / "output_files" / "k_points.out.gz"
    parser = StdoutParser(file_name)
    results = parser.parse()
    assert [0.25, -0.25, 0.333333] in results["input"]["k_points"]
    assert 0.04166667 in results["input"]["k_point_weights"]


def test_metadata(data_dir):
    file_name = data_dir / "stdout" / "output_files" / "Si.output_cbm_vbm.out"
    parser = StdoutParser(file_name)
    _ = parser.parse()
    assert parser.run_metadata["xc"] == "pbesol"
    assert parser.run_metadata["num_tasks"] == 4
    assert parser.run_metadata["num_electrons"] == 28.0
    assert parser.run_metadata["num_bands"] == 24
    assert parser.run_metadata["relax"] is False


def test_stdout_parser_regression(data_dir):
    """Checks parsing given output files against reference values."""
    stdout_data_dir = data_dir / "stdout"
    for file_name in stdout_data_dir.iterdir():
        if file_name.is_file():
            parser = StdoutParser(file_name)
            results = parser.parse()
            with open(stdout_data_dir / "ref" / (file_name.stem + ".yaml")) as f:
                ref_data = yaml.safe_load(f)
            assert is_subset(ref_data, results)


# def test_print_parsed_results(data_dir):
#     """This test is for development purposes only."""
#     file_name = data_dir / "stdout" / "output_files" / "Si.output_cbm_vbm.out"
#     parser = StdoutParser(file_name)
#     results = parser.parse()
#     from pprint import pprint
#     pprint(parser.run_metadata)



"""Electronic configurations for the FHI-aims species' defaults."""

from dataclasses import dataclass, field
from enum import Enum

from pyfhiaims.control.chunk import AimsControlChunk


class Orbital(Enum):
    """Enum mapping l quantum number to letter."""

    S = 0
    P = 1
    D = 2
    F = 3
    G = 4
    H = 5
    I = 6


class ElectronicConfigurationType(Enum):
    """Enumb mapping valence and ion occ orbitals."""

    VALENCE = "valence"
    IONIC = "ion_occ"


@dataclass
class ElectronicConfiguration(AimsControlChunk):
    """Electronic configuration, both valence and ionic."""

    type: ElectronicConfigurationType = None
    occupations: dict[(int, Orbital):int] = field(default_factory=dict)

    @classmethod
    def from_strings(cls, config_str) -> "ElectronicConfiguration":
        """Get the electronic configuration from a multi-line string."""
        config_type = None
        config = {}
        for line in config_str:
            type_str, n, l, occ = line.split()
            if config_type is None:
                config_type = ElectronicConfigurationType(type_str)
            elif config_type != ElectronicConfigurationType(type_str):
                raise ValueError("config_type is inconsistent with type_str")
            config[(int(n), Orbital[l.upper()])] = int(float(occ))
        return cls(
            type=config_type,
            occupations=config,
        )

    @property
    def n_electrons(self) -> int:
        """The number of electrons in the given configuration."""
        num = 0
        for k, v in self.occupations.items():
            n, l = k
            num += (n - l.value - 1) * (4 * l.value + 2) + v
        return num

    def to_string(self) -> str:
        """Get the string representation of the configuration."""
        return "\n".join(
            [
                f"{self.type.value:>11} {k[0]:6d}  {k[1].name.lower()} {v:3d}."
                for k, v in self.occupations.items()
            ]
        )


class ValenceElectronicConfiguration(ElectronicConfiguration):
    """Valence electron basis function definition."""

    type = ElectronicConfigurationType.VALENCE
    keywords = ("valence",)


class IonicElectronicConfiguration(ElectronicConfiguration):
    """The ionic occupation basis function definition."""

    type = ElectronicConfigurationType.IONIC
    keywords = ("ion_occ",)



"""Classes for reading/manipulating/writing FHI-aims control.in files."""

from __future__ import annotations

import re
import time
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING
from warnings import warn

import numpy as np
from monty.json import MontyDecoder, MSONable

from pyfhiaims.errors import PyaimsError
from pyfhiaims.geometry.geometry import AimsGeometry
from pyfhiaims.species_defaults.species import SpeciesDefaults

if TYPE_CHECKING:
    from typing import Any, Self, TextIO

    from ase import Atoms
    from pymatgen.core import Molecule, Structure

    from pyfhiaims.control.kpoints import AimsKPoints

__author__ = "Thomas A. R. Purcell"
__version__ = "1.0"
__email__ = "purcellt@arizona.edu"
__date__ = "July 2024"


@dataclass
class AimsControl(MSONable):
    """An FHI-aims control.in file.

    Attributes:
        parameters (dict[str, Any]): The parameters' dictionary containing all input
            flags (key) and values for the control.in file

    """

    parameters: dict[str, Any] = field(default_factory=dict)
    outputs: list[str] = field(default_factory=list)
    k_points: AimsKPoints = None
    species_defaults: dict[str, SpeciesDefaults] = field(default_factory=dict)  # None?

    def __getitem__(self, key: str) -> Any:
        """Get an input parameter.

        Args:
            key (str): The parameter to get

        Returns:
            The setting for that parameter

        Raises:
            KeyError: If the key is not in self._parameters

        """
        if key == "output":
            return self.outputs

        if key not in self.parameters:
            raise KeyError(f"{key} not set in AimsControl")
        return self.parameters[key]

    def __setitem__(self, key: str, value: Any) -> None:
        """Set an attribute of the class.

        Args:
            key (str): The parameter to get
            value (Any): The value for that parameter

        """
        if key == "output":
            warn(
                "Outputs are set seperately, use the outputs property",
                RuntimeWarning,
                stacklevel=1,
            )
        else:
            self.parameters[key] = value

    def __delitem__(self, key: str) -> Any:
        """Delete a parameter from the input object.

        Args:
        key (str): The key in the parameter to remove

        Returns:
            Either the value of the deleted parameter or None if key is
            not in self._parameters

        """
        if key == "output":
            self.outputs = []

        return self.parameters.pop(key, None)

    @staticmethod
    def get_aims_control_parameter_str(key: str, value: Any, fmt: str) -> str:
        """Get the string needed to add a parameter to the control.in file.

        Args:
            key (str): The name of the input flag
            value (Any): The value to be set for the flag
            fmt (str): The format string to apply to the value

        Returns:
            str: The line to add to the control.in file

        """
        if value is None:
            return ""
        return f"{key:50s}{fmt % value}\n"

    def get_content(
        self,
        geometry: AimsGeometry | Atoms | Structure | Molecule,
        verbose_header: bool = False,
    ) -> str:
        """Get the content of the file.

        Args:
            geometry: The geometry to write the input file for
            verbose_header (bool): If True print the input option dictionary

        Returns:
            str: The content of the file for a given geometry

        """
        parameters = deepcopy(self.parameters)

        if geometry.__class__.__name__ == "Atoms":
            geometry = AimsGeometry.from_atoms(geometry)
        elif geometry.__class__.__name__ in ("Structure", "Molecule"):
            geometry = AimsGeometry.from_structure(geometry)

        lim = "#" + "=" * 79
        content = ""
        cubes = parameters.pop("cubes", [])

        if verbose_header:
            content += "#\n# List of parameters used to initialize the calculator:\n"
            for param, val in parameters.items():
                content += f"#     {param}:{val}\n"
            content += f"#     output:{self.outputs}\n"
            content += f"{lim}\n"

        if parameters["xc"] == "LDA":
            parameters["xc"] = "pw-lda"

        tiers = parameters.pop("tier", None)
        plus_u = parameters.pop("plus_u", None)
        species_dir = parameters.pop("species_dir", None)

        if isinstance(tiers, int):
            tiers = dict.fromkeys(np.unique(geometry.symbols), tiers)
        elif tiers is not None:
            assert all(sym in tiers for sym in np.unique(geometry.symbols))

        # Don't override species defaults
        if species_dir is not None and Path(species_dir).exists():
            geometry.load_species(species_directory=species_dir)

        for sym in np.unique(geometry.symbols):
            if tiers is not None:
                end_activate = 1 + min(
                    tiers[sym], geometry.species_dict[sym].basis_set.n_tiers
                )
                for tt in range(1, end_activate):
                    geometry.species_dict[sym].basis_set.activate_tier(tt)

                for tt in range(
                    end_activate, geometry.species_dict[sym].basis_set.n_tiers + 1
                ):
                    geometry.species_dict[sym].basis_set.deactivate_tier(tt)

            if plus_u is not None:
                geometry.species_dict[sym].plus_u = plus_u.get(sym)

        if all(inp in parameters for inp in ["smearing", "occupation_type"]):
            raise ValueError(
                "Both smearing and occupation_type can't be "
                "in the same parameters file."
            )

        for key, value in parameters.items():
            if key == "smearing":
                name = parameters["smearing"][0].lower()
                if name == "fermi-dirac":
                    name = "fermi"
                width = parameters["smearing"][1]
                if name == "methfessel-paxton":
                    order = parameters["smearing"][2]
                    order = f"{order:d}"
                else:
                    order = ""

                content += self.get_aims_control_parameter_str(
                    "occupation_type", (name, width, order), "%s %f %s"
                )
            elif key == "vdw_correction_hirshfeld" and value:
                content += self.get_aims_control_parameter_str(key, "", "%s")
            elif key == "xc":
                if "libxc" in value:
                    warn(
                        f"Not all libxc functionals will work with FHI-aims "
                        f"be careful when using {value}",
                        stacklevel=1,
                    )
                    content += self.get_aims_control_parameter_str(
                        "override_warning_libxc", ".true.", "%s"
                    )
                content += self.get_aims_control_parameter_str(key, value, "%s")
            elif isinstance(value, bool):
                content += self.get_aims_control_parameter_str(
                    key, str(value).lower(), ".%s."
                )
            elif isinstance(value, tuple | list):
                content += self.get_aims_control_parameter_str(
                    key, " ".join(map(str, value)), "%s"
                )
            elif isinstance(value, str):
                content += self.get_aims_control_parameter_str(key, value, "%s")
            else:
                content += self.get_aims_control_parameter_str(key, value, "%r")

        for output_type in self.outputs:
            content += self.get_aims_control_parameter_str("output", output_type, "%s")

        if cubes:
            for cube in cubes:
                content += cube.control_block

        content += f"{lim}\n\n"

        for sp in geometry.species_dict.values():
            content += sp.content

        return content

    def write_file(
        self,
        geometry: AimsGeometry,
        writer: str | Path | TextIO | None = None,
        header: str | None = None,
        verbose_header: bool = False,
        overwrite: bool = False,
    ) -> None:
        """Write the control.in file.

        Args:
            geometry (AimsGeometry): The structure to write the input
                file for
            writer (str | Path | TextIO): The directory to write the control.in file.
                If None use cwd
            header (str | None): An optional header to add to the file
            verbose_header (bool): If True print the input option dictionary
            overwrite (bool): If True allow to overwrite existing files

        Raises:
            ValueError: If a file must be overwritten and overwrite is False
            ValueError: If k-grid is not provided for the periodic structures

        """
        writer = writer or Path.cwd()
        if isinstance(writer, str | Path):
            if (Path(writer) / "control.in").exists() and not overwrite:
                raise ValueError(f"control.in file already in {writer}")
            fname = f"{writer}/control.in"
        else:
            try:
                fname = writer.name
            except AttributeError:
                fname = "control.in"
        if header is None:
            header = "\n".join(
                [
                    f"#{'=' * 79}\n",
                    f"# FHI-aims control file: {fname}\n",
                    f"# {time.asctime()}\n",
                    f"#{'=' * 79}\n",
                ]
            )

        if (geometry.lattice_vectors is not None) and (
            "k_grid" not in self.parameters and "k_grid_density" not in self.parameters
        ):
            raise ValueError("k-grid must be defined for periodic systems")

        content = header + self.get_content(geometry, verbose_header)

        if isinstance(writer, str | Path):
            with open(fname, mode="w") as file:
                file.write(content)
        else:
            writer.write(content)

    def as_dict(self) -> dict[str, Any]:
        """Get a dictionary representation of the control.in file."""
        dct: dict[str, Any] = {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "parameters": self.parameters,
            "outputs": self.outputs,
        }
        return dct

    @classmethod
    def from_strings(csl, content_str: str) -> AimsControl:
        """Generate control.in file from its content str."""
        parameters = {}
        outputs = []

        # get species' defaults from the file first
        species = {}
        species_re = re.compile(r"(?<=\n) *species.*?(?=\n *species|$)", re.DOTALL)
        species_lines = re.findall(species_re, content_str)
        element_re = re.compile(r" *species *(\S+)", re.DOTALL)

        for block in species_lines:
            if re.match(element_re, block) is not None:
                element = re.match(element_re, block).group(1)
                species[element] = SpeciesDefaults.from_strings(block.split("\n"))

        # then everything else
        for line in content_str.splitlines():
            # remove comments and blank lines
            idx = line.find("#")
            line = line.strip() if idx == -1 else line[:idx].strip()
            if not line:
                continue
            # stop at species
            if line.startswith("species "):
                break
            k, v = line.split(maxsplit=1)
            if k == "output":
                if "cube" in v:
                    # TODO: does not work with Cubes yet
                    raise PyaimsError(
                        "Reading control.in from file does not work with cubes yet"
                    )
                outputs.append(v)
            else:
                parameters[k] = v

        return AimsControl(
            parameters=parameters, outputs=outputs, species_defaults=species
        )

    @classmethod
    def from_file(cls, control_file: str | Path) -> AimsControl:
        """Instantiate the Control object."""
        with open(control_file) as f:
            lines = f.read()

        return cls.from_strings(lines)

    @classmethod
    def from_dict(cls, dct: dict[str, Any]) -> Self:
        """Initialize from dictionary.

        Args:
            dct (dict[str, Any]): The MontyEncoded dictionary

        Returns:
            The AimsControl for dct

        """
        decoded = {
            key: MontyDecoder().process_decoded(val)
            for key, val in dct.items()
            if not key.startswith("@")
        }

        return cls(parameters=decoded["parameters"], outputs=decoded["outputs"])



"""Definition of the Basis Function class."""

from __future__ import annotations

import inspect
import sys
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence


class BasisFunctionError(Exception):
    """Exception raised if an error occurs when parsing an Aims output file."""


@dataclass(kw_only=True)
class BasisFunction:
    """Defines an FHI-aims basis function.

    Parameters
    ----------
    n: int
        Radial quantum number
    l: str
        The angular momentum of the function
    type: str
        A type of basis function (from `hydro`, `ionic`, `confined`, `sto`, `gaussian`)
    enabled: bool
        A flag indicating if the basis function is enabled (not commented out
    aux: bool
        A flag indicating if the basis function belongs to the auxiliary basis set used
        to expand Coulomb operator

    """

    n: int
    l: str | int
    type: str
    enabled: bool = True
    aux: bool = False

    def content_block(self, content_fields: Sequence[str]) -> str:
        """Get the content block for the control.in file."""
        content = [str(getattr(self, f)) for f in content_fields]
        return (
            f"{'#' if not self.enabled else ' '} "
            f"{('for_aux' if self.aux else ''):>7}{self.type:>9} "
            f"{self.n:6d}  {self.l}  {' '.join(content)}"
        )

    @classmethod
    def from_string(cls, string: str) -> BasisFunction:
        """Get the general BasisFunction object from string representation."""
        string = string.strip()
        enabled = not string.startswith("#")
        if not enabled:
            string = string[1:]
        for_aux = "aux" in string
        if for_aux:
            if "gaussian" in string:
                string = string.replace("aux_", "")
            else:
                string = string.replace("for_aux", "")

        for k, v in BASIS_TYPES.items():
            if k in string:
                basis_function = v.from_string(string)
                basis_function.enabled = enabled
                basis_function.aux = for_aux
                return basis_function
        raise ValueError(f"No valid basis function type found in {string}")


@dataclass(kw_only=True)
class HydroBasisFunction(BasisFunction):
    """The Hydrogen like basis function for FHI-aims.

    Parameters
    ----------
    z_eff: float
        scales the radial function as an effective nuclear charge in the defining
        Coulomb potential zeff/r.

    """

    z_eff: float
    type: str = "hydro"

    @classmethod
    def from_string(cls, string: str) -> HydroBasisFunction:
        """Get HydroBasisFunction object from string representation."""
        _, n, l, z_eff = string.split()
        return cls(n=int(n), l=l, z_eff=float(z_eff))

    def __str__(self) -> str:
        """Get the string representation for the control.in file."""
        return self.content_block(["z_eff"])


@dataclass(kw_only=True)
class IonicBasisFunction(BasisFunction):
    """The ionic basis function for FHI-aims.

    Parameters
    ----------
    radius: float
        The onset radius of the confining potential
        (in atomic units, 1 a.u.= 0.529177 Å). If the word auto is specified instead
        of a numerical value, the default onset radius given in the cut_pot
        tag is used.

    """

    radius: float | str
    type: str = "ionic"

    @classmethod
    def from_string(cls, string: str) -> IonicBasisFunction:
        """Get IonicBasisFunction object from string representation."""
        _, n, l, radius = string.split()
        return cls(
            n=int(n),
            l=l,
            radius="auto" if radius == "auto" else float(radius),
        )

    def __str__(self) -> str:
        """Get the string representation for the control.in file."""
        return self.content_block(["radius"])


@dataclass(kw_only=True)
class ConfinedBasisFunction(BasisFunction):
    """The confined free-atom like radial basis function.

    Parameters
    ----------
    radius: float | str
        The onset radius of the confining potential
        (in atomic units, 1 a.u.= 0.529177 Å). If the word auto is specified instead
        of a numerical value, the default onset radius given in the cut_pot
        tag is used.

    """

    radius: float | str
    type: str = "confined"

    @classmethod
    def from_string(cls, string: str) -> ConfinedBasisFunction:
        """Get ConfinedBasisFunction object from string representation."""
        enabled = not string.startswith("#")
        if not enabled:
            string = string[1:].lstrip()

        _, n, l, radius = string.split()
        return cls(
            n=int(n),
            l=l,
            radius="auto" if radius == "auto" else float(radius),
            enabled=enabled,
        )

    def __str__(self) -> str:
        """Get the string representation for the control.in file."""
        return self.content_block(["radius"])


@dataclass(kw_only=True)
class STOBasisFunction(BasisFunction):
    """The STO basis function for FHI-aims.

    Parameters
    ----------
    zeta: float
        The STO exponent, which plays the role of the effective nuclear charge

    """

    zeta: float = field(default_factory=float)
    type: str = "sto"

    @classmethod
    def from_string(cls, string: str) -> STOBasisFunction:
        """Get ConfinedBasisFunction object from string representation."""
        enabled = not string.startswith("#")
        if not enabled:
            string = string[1:].lstrip()

        _, n, l, zeta = string.split()
        l = int(l)
        return cls(n=int(n), l=l, zeta=float(zeta), enabled=enabled)

    def __str__(self) -> str:
        """Get the string representation for the control.in file."""
        return self.content_block(["zeta"])


@dataclass(kw_only=True)
class GaussianBasisFunction(BasisFunction):
    """The Gaussian basis function for FHI-aims.

    Parameters
    ----------
    n: int
        The number of Gaussian functions to comprise the basis function
    alpha_i: Sequence[float]
        The exponent defining a primitive Gaussian function [in bohr−2].
    coeff_i: Optional[Sequence[float]]
        The coefficients for the Gaussian functions if n > 1

    """

    alpha_i: Sequence[float]
    coeff_i: Sequence[float] | None
    type: str = "gaussian"

    def __post_init__(self):
        """Post init checks."""
        if self.n > 1 and not self.coeff_i:
            raise BasisFunctionError(
                f"Total number of Gaussians is {self.n} so coeff_i must be passed."
            )

        if len(self.alpha_i) != self.n:
            raise BasisFunctionError(f"The length of alpha_i must be {self.n}.")

        if self.coeff_i and (len(self.coeff_i) != self.n):
            raise BasisFunctionError(f"The length of coeff_i must be {self.n}.")

    @classmethod
    def from_string(cls, string: str) -> GaussianBasisFunction:
        """Get GaussianBasisFunction object from string representation."""
        alpha_i = []
        coeff_i = []
        lines = string.splitlines()
        _, l, *n = lines[0].split()
        if int(n[0]) == 1:
            alpha_i.append(float(n[1]))
        else:
            for line in lines[1:]:
                alpha_i.append(float(line.replace("#", "").strip().split()[0]))
                coeff_i.append(float(line.replace("#", "").strip().split()[1]))

        return cls(
            n=int(n[0]),
            l=l,
            alpha_i=alpha_i,
            coeff_i=coeff_i,
        )

    def __str__(self) -> str:
        """Get the content block for the control.in file."""
        title_line = (
            f"{'#' if not self.enabled else ''}"
            f"{'aux_' if self.aux else ''}{self.type:<11} "
            f"{self.l}  {self.n:6d}"
        )
        if self.n == 1:
            return f"{title_line}  {self.alpha_i[0]:.7f}"

        lines = [
            title_line,
        ]
        for alpha, coeff in zip(self.alpha_i, self.coeff_i, strict=False):
            lines.append(
                f"{'#' if not self.enabled else ' '}   {alpha:.7f}  {coeff:.7f}"
            )

        return "\n".join(lines)


BASIS_TYPES = {
    cls.type: cls
    for _, cls in inspect.getmembers(
        sys.modules[__name__],
        lambda member: inspect.isclass(member)
        and issubclass(member, BasisFunction)
        and member.__name__ != "BasisFunction",
    )
}



import numpy as np
import pytest

from pyfhiaims.geometry.atom import ATOMIC_SYMBOLS_TO_NUMBERS
from pyfhiaims.geometry.geometry import AimsGeometry, InvalidGeometryError, singular
from pyfhiaims.species_defaults.species import SpeciesDefaults
from tests import DATA_DIR

GEO_DATA_DIR = DATA_DIR / "geometry"

h2o_content = """atom         0.000000000000e+00   0.000000000000e+00   0.000000000000e+00 O
    magnetic_response
    nuclear_spin 1.000000000000
    isotope 17.0
atom         7.070000000000e-01  -7.070000000000e-01   0.000000000000e+00 H
    magnetic_response
    magnetic_moment 0.500000000000
atom        -7.070000000000e-01  -7.070000000000e-01   0.000000000000e+00 H
empty       -1.000000000000e+00   0.000000000000e+00   0.000000000000e+00 O
pseudocore   1.000000000000e+00   0.000000000000e+00   0.000000000000e+00 O"""  # noqa: E501

scr_content = """lattice_vector -2.612201390000000e+00 2.612201390000000e+00 5.091086090000000e+00
lattice_vector 2.612201390000000e+00 -2.612201390000000e+00 5.091086090000000e+00
lattice_vector 2.612201390000000e+00 2.612201390000000e+00 -5.091086090000000e+00
atom_frac    5.000000000000e-01   5.000000000000e-01  -5.000000000000e-03 Al
atom_frac    7.500000000000e-01   2.500000000000e-01   5.000000000000e-01 Al
atom_frac    1.604770860000e+00   1.622146130000e+00   1.491642500000e+00 S
atom_frac    1.305036200000e-01   1.131283600000e-01  -1.491642500000e+00 S
atom_frac   -1.622146130000e+00  -1.305036200000e-01  -1.737527000000e-02 S
atom_frac   -1.131283600000e-01  -1.604770860000e+00   1.737527000000e-02 S
atom_frac    0.000000000000e+00  -0.000000000000e+00   0.000000000000e+00 Zn
symmetry_n_params 5 2 3
symmetry_params a c x4 y4 z4
symmetry_lv -(1.0/2.0)*a , (1.0/2.0)*a , (1.0/2.0)*c
symmetry_lv (1.0/2.0)*a , -(1.0/2.0)*a , (1.0/2.0)*c
symmetry_lv (1.0/2.0)*a , (1.0/2.0)*a , -(1.0/2.0)*c
symmetry_frac (1.0/2.0) , (1.0/2.0) , 0.0
symmetry_frac (3.0/4.0) , (1.0/4.0) , (1.0/2.0)
symmetry_frac (y4+z4) , (z4+x4) , (x4+y4)
symmetry_frac (z4-y4) , (z4-x4) , -(x4+y4)
symmetry_frac -(z4+x4) , (y4-z4) , (y4-x4)
symmetry_frac (x4-z4) , -(y4+z4) , (x4-y4)
symmetry_frac 0.0 , 0.0 , 0.0"""  # noqa: E501

mg2mn4o8_content = """lattice_vector 5.068823430000000e+00 1.248800000000000e-04 -2.661101670000000e+00
lattice_vector -1.397042340000000e+00 4.872499110000000e+00 -2.661102030000000e+00
lattice_vector 9.860910000000001e-03 1.308528000000000e-02 6.176493590000000e+00
    constrain_relaxation z
atom_frac    3.748972600000e-01   6.251027400000e-01   7.500000200000e-01 Mg
atom_frac    6.251027400000e-01   3.748972600000e-01   2.499999800000e-01 Mg
atom_frac   -2.692415593343e-19  -3.560208392073e-19   5.000000000000e-01 Mn
    initial_charge 0.500000000000
    initial_moment 4.500000000000
atom_frac    0.000000000000e+00   5.000000000000e-01   0.000000000000e+00 Mn
    initial_moment 5.000000000000
atom_frac    5.000000000000e-01   0.000000000000e+00   5.000000000000e-01 Mn
    initial_moment 5.000000000000
atom_frac    0.000000000000e+00   0.000000000000e+00   0.000000000000e+00 Mn
    initial_moment 5.000000000000
atom_frac    7.540230900001e-01   7.782675000001e-01   5.080588200001e-01 O
    velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00
atom_frac    7.702028500000e-01   2.459477900001e-01   9.919131600001e-01 O
    RT_TDDFT_initial_velocity   1.000000000000e+01   2.000000000000e+01   3.000000000000e+01
atom_frac    2.217325400000e-01   2.459768899999e-01   9.919411600000e-01 O
atom_frac    2.459769100000e-01   2.217324999999e-01   4.919411800000e-01 O
    constrain_relaxation x y
    constraint_region 1
atom_frac    2.459476500000e-01   7.702028800001e-01   4.919131300000e-01 O
    constrain_relaxation x y z
atom_frac    2.297971500000e-01   7.540522099999e-01   8.086839999943e-03 O
atom_frac    7.540523500000e-01   2.297971199999e-01   5.080868700000e-01 O
atom_frac    7.782674600000e-01   7.540231100001e-01   8.058839999985e-03 O"""  # noqa: E501


def test_h2o():
    geometry = AimsGeometry.from_file(f"{GEO_DATA_DIR}/h2o.in")
    positions = np.array(
        [
            [0.0, 0.0, 0.0],
            [0.707, -0.707, 0.0],
            [-0.707, -0.707, 0.0],
            [-1.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
        ]
    )
    all_none_atom_attrs = [
        "velocities",
        "RT_TDDFT_initial_velocities",
        "nuclear_constraints",
        "constraint_regions",
    ]
    geo_none_attrs = [
        "lattice_vectors",
        "lattice_constraints",
        "hessian_block",
        "hessian_block_lv",
        "hessian_block_lv_atoms",
        "hessian_file",
        "trust_radius",
        "symmetry_n_params",
        "symmetry_params",
        "symmetry_lv",
        "symmetry_frac",
        "symmetry_frac_change_threshold",
        "homogeneous_field",
        "multipole",
        "esp_constraint",
        "calculate_friction",
    ]
    assert geometry.n_atoms == 5
    assert geometry.symbols == ["O", "H", "H", "O", "O"]
    assert geometry.numbers == [8, 1, 1, 8, 8]
    with pytest.raises(InvalidGeometryError):
        geometry._species_property("mass")
    assert np.allclose(geometry.positions, positions)
    assert np.allclose(geometry.initial_charges, np.zeros(5))
    assert np.allclose(geometry.initial_moments, np.zeros(5))
    assert geometry.is_empty_atoms == [False, False, False, True, False]
    assert geometry.is_pseudocore_atoms == [False, False, False, False, True]
    assert geometry.magnetic_responses == [True, True, None, None, None]
    assert geometry.nuclear_spins == [1.0, None, None, None, None]
    assert geometry.isotopes == [17, None, None, None, None]
    assert geometry.magnetic_moments == [None, 0.5, None, None, None]

    for attr in all_none_atom_attrs:
        assert all([val is None for val in getattr(geometry, attr)])  # noqa: C419

    for attr in geo_none_attrs:
        assert getattr(geometry, attr) is None

    assert geometry.to_string() == h2o_content


def test_scr(data_dir):
    geometry = AimsGeometry.from_file(f"{GEO_DATA_DIR}/scr.in")
    positions = np.array(
        [
            [0.50000000, 0.50000000, -0.00500000],
            [0.75000000, 0.25000000, 0.50000000],
            [1.60477086, 1.62214613, 1.49164250],
            [0.13050362, 0.11312836, -1.49164250],
            [-1.62214613, -0.13050362, -0.01737527],
            [-0.11312836, -1.60477086, 0.01737527],
            [0.00000000, 0.00000000, 0.00000000],
        ]
    )
    symmetry_lv = [
        ["-(1.0/2.0)*a", "(1.0/2.0)*a", "(1.0/2.0)*c"],
        ["(1.0/2.0)*a", "-(1.0/2.0)*a", "(1.0/2.0)*c"],
        ["(1.0/2.0)*a", "(1.0/2.0)*a", "-(1.0/2.0)*c"],
    ]
    symmetry_frac = [
        ["(1.0/2.0)", "(1.0/2.0)", "0.0"],
        ["(3.0/4.0)", "(1.0/4.0)", "(1.0/2.0)"],
        ["(y4+z4)", "(z4+x4)", "(x4+y4)"],
        ["(z4-y4)", "(z4-x4)", "-(x4+y4)"],
        ["-(z4+x4)", "(y4-z4)", "(y4-x4)"],
        ["(x4-z4)", "-(y4+z4)", "(x4-y4)"],
        ["0.0", "0.0", "0.0"],
    ]
    all_none_atom_attrs = [
        "velocities",
        "RT_TDDFT_initial_velocities",
        "nuclear_constraints",
        "constraint_regions",
        "magnetic_responses",
        "magnetic_moments",
        "nuclear_spins",
        "isotopes",
    ]
    geo_none_attrs = [
        "hessian_block",
        "hessian_block_lv",
        "hessian_block_lv_atoms",
        "hessian_file",
        "trust_radius",
        "homogeneous_field",
        "multipole",
        "esp_constraint",
        "calculate_friction",
    ]
    assert geometry.symbols == ["Al", "Al", "S", "S", "S", "S", "Zn"]
    assert geometry.numbers == [
        ATOMIC_SYMBOLS_TO_NUMBERS[sym] for sym in ["Al", "Al", "S", "S", "S", "S", "Zn"]
    ]
    assert np.allclose(geometry.fractional_positions, positions)
    assert np.allclose(geometry.initial_charges, np.zeros(7))
    assert np.allclose(geometry.initial_moments, np.zeros(7))
    assert not any(geometry.is_empty_atoms)
    assert not any(geometry.is_pseudocore_atoms)
    assert not any(np.array(geometry.lattice_constraints).flatten())
    for attr in all_none_atom_attrs:
        assert all([val is None for val in getattr(geometry, attr)])  # noqa: C419

    for attr in geo_none_attrs:
        assert getattr(geometry, attr) is None

    assert geometry.symmetry_n_params == (5, 2, 3)
    assert geometry.symmetry_params == ["a", "c", "x4", "y4", "z4"]
    assert geometry.symmetry_lv == symmetry_lv
    assert geometry.symmetry_frac == symmetry_frac
    assert geometry.to_string() == scr_content

    geometry.load_species(f"{data_dir}/species_dir/")
    al_species = SpeciesDefaults.from_file(f"{data_dir}/species_dir/13_Al_default")
    s_species = SpeciesDefaults.from_file(f"{data_dir}/species_dir/16_S_default")
    zn_species = SpeciesDefaults.from_file(f"{data_dir}/species_dir/30_Zn_default")

    assert np.all(geometry._species_property("nucleus") == geometry.nuclear_charges)
    assert np.all(geometry._species_property("mass") == geometry.masses)
    assert geometry.get_species("S").content == s_species.content
    assert geometry.get_species("Zn").content == zn_species.content
    assert geometry.get_species("Al").content == al_species.content

    assert s_species.content in geometry.species_block
    assert zn_species.content in geometry.species_block
    assert al_species.content in geometry.species_block

    geometry.species_dict.pop("Al")
    with pytest.raises(
        InvalidGeometryError,
        match="Species are not defined for all atoms in the structure"
    ):
        _ = geometry.species_block

    with pytest.raises(
        InvalidGeometryError,
        match="Lattice vectors must be defined when"
    ):
        _ = AimsGeometry(geometry.atoms, symmetry_frac=geometry.symmetry_frac)

    with pytest.raises(InvalidGeometryError, match="must be 3x3 not"):
        _ = AimsGeometry(
            geometry.atoms,
            geometry.lattice_vectors,
            symmetry_n_params=geometry.symmetry_n_params,
            symmetry_params=geometry.symmetry_params,
            symmetry_frac=geometry.symmetry_frac,
            symmetry_lv=[*geometry.symmetry_lv, ["a", "b", "c"]],
        )


def test_mg2mn4o8():
    geometry = AimsGeometry.from_file(f"{GEO_DATA_DIR}/mg2mn4o8.in")
    positions = np.array(
        [
            [1.034388703398e00, 3.055673321740e00, 1.971268420993e00],
            [2.647253296602e00, 1.830035948260e00, -1.116978530993e00],
            [4.930455000000e-03, 6.542640000000e-03, 3.088246795000e00],
            [-6.985211700000e-01, 2.436249555000e00, -1.330551015000e00],
            [2.539342170000e00, 6.605080000000e-03, 1.757695960000e00],
            [0.000000000000e00, 0.000000000000e00, 0.000000000000e00],
            [2.739747178306e00, 3.798849955412e00, -9.395592830776e-01],
            [3.570203942272e00, 1.211456032248e00, 3.422465020960e00],
            [7.900644064579e-01, 1.211529695387e00, 4.882095782598e00],
            [9.418948216942e-01, 1.086859314588e00, 1.793849173078e00],
            [1.755098882262e-01, 3.759280382304e00, 3.342181446544e-01],
            [1.114380577281e-01, 3.674253237752e00, -2.568175130960e00],
            [3.506132111774e00, 1.126428887696e00, 5.200717453456e-01],
            [2.891577593542e00, 3.674179574613e00, -4.027805892598e00],
        ]
    )

    symbols = [
        "Mg",
        "Mg",
        "Mn",
        "Mn",
        "Mn",
        "Mn",
        "O",
        "O",
        "O",
        "O",
        "O",
        "O",
        "O",
        "O",
    ]
    initial_charges = np.zeros(14)
    initial_charges[2] = 0.5
    initial_moments = np.zeros(14)
    initial_moments[2:6] = [4.5, 5, 5, 5]
    constraint_regions = 14 * [None]
    constraint_regions[-5] = 1
    nuclear_constraints = 14 * [None]
    nuclear_constraints[-5] = [True, True, False]
    nuclear_constraints[-4] = [True, True, True]
    RT_TDDFT_initial_velocities = 14 * [None]
    RT_TDDFT_initial_velocities[-7] = [10.0, 20.0, 30.0]
    velocities = 14 * [None]
    velocities[-8] = [1.0, 2.0, 3.0]
    all_none_atom_attrs = [
        "magnetic_responses",
        "magnetic_moments",
        "nuclear_spins",
        "isotopes",
    ]
    geo_none_attrs = [
        "hessian_block",
        "hessian_block_lv",
        "hessian_block_lv_atoms",
        "hessian_file",
        "trust_radius",
        "symmetry_n_params",
        "symmetry_params",
        "symmetry_lv",
        "symmetry_frac",
        "symmetry_frac_change_threshold",
        "homogeneous_field",
        "multipole",
        "esp_constraint",
        "calculate_friction",
    ]
    assert geometry.symbols == symbols
    assert geometry.numbers == [ATOMIC_SYMBOLS_TO_NUMBERS[sym] for sym in symbols]
    assert np.allclose(geometry.positions, positions, 1e-08)
    assert np.allclose(geometry.initial_charges, initial_charges)
    assert np.allclose(geometry.initial_moments, initial_moments)
    assert geometry.constraint_regions == constraint_regions
    for test, ref in zip(
        geometry.nuclear_constraints, nuclear_constraints, strict=False
    ):
        assert test == ref
    for test, ref in zip(
        geometry.RT_TDDFT_initial_velocities, RT_TDDFT_initial_velocities, strict=False
    ):
        assert test == ref
    for test, ref in zip(geometry.velocities, velocities, strict=False):
        assert test == ref

    assert np.allclose(
        geometry.lattice_vectors,
        np.array(
            [
                [5.068823430000e00, 1.248800000000e-04, -2.661101670000e00],
                [-1.397042340000e00, 4.872499110000e00, -2.661102030000e00],
                [9.860910000000e-03, 1.308528000000e-02, 6.176493590000e00],
            ]
        ),
    )
    assert np.all(
        np.array(geometry.lattice_constraints)
        == np.array(
            [[False, False, False], [False, False, False], [False, False, True]]
        )
    )

    for attr in all_none_atom_attrs:
        assert all([val is None for val in getattr(geometry, attr)])  # noqa: C419

    assert not any(geometry.is_empty_atoms)
    assert not any(geometry.is_pseudocore_atoms)

    for attr in geo_none_attrs:
        assert getattr(geometry, attr) is None

    assert geometry.to_string() == mg2mn4o8_content

    with pytest.raises(InvalidGeometryError):
        _ = AimsGeometry(
            geometry.atoms,
            lattice_vectors=np.eye(4),
        )

    with pytest.raises(InvalidGeometryError):
        _ = AimsGeometry(
            geometry.atoms,
            lattice_vectors=np.eye(3) * 0.0,
        )

    with pytest.raises(InvalidGeometryError):
        _ = AimsGeometry(
            geometry.atoms,
            lattice_vectors=np.eye(3) * 0.0,
        )

    with pytest.raises(InvalidGeometryError):
        _ = AimsGeometry(
            geometry.atoms,
            lattice_constraints=[
                [True, True, True],
                [True, True, True],
                [True, True, True],
            ],
        )


def test_from_strings(tmp_path):
    """Test making a geometry file from strings"""
    geometry = AimsGeometry.from_strings(scr_content.splitlines())
    geometry.hessian_file = "hessian.aims"
    geometry.symmetry_frac_change_threshold = 0.1
    geometry.trust_radius = 0.2
    geometry.homogeneous_field = [0.1, 0.2, 0.3]
    geometry.multipole = [(1.0, 2.1, 3.2, 1, 4.3)]
    geometry.esp_constraint = (1.1, 2.3)
    geometry.verbatim_writeout = True
    geometry.calculate_friction = True
    geometry.hessian_block = [(1, 1, np.eye(3))]
    geometry.hessian_block_lv = [(1, 2, np.eye(3) * 2)]
    geometry.hessian_block_lv_atoms = [(2, 1, np.eye(3) * 1.2)]

    geometry.write_file(f"{tmp_path}/geometry.in")

    geo_2 = AimsGeometry.from_strings(geometry.to_string().splitlines())
    assert geo_2.to_string() == geometry.to_string()

    geo_2 = AimsGeometry.from_file(f"{tmp_path}/geometry.in")
    assert geometry.to_string() == geo_2.to_string()

    geometry.homogeneous_field = [0.0, 1.0]
    with pytest.raises(
        InvalidGeometryError,
        match="The provided homogeneous_field value is invalid"
    ):
        _ = geometry.to_string()
    geometry.homogeneous_field = None

    geometry.multipole = [(0, 0, 0)]
    with pytest.raises(
        InvalidGeometryError,
        match="th multipole value is invalid"

    ):
        _ = geometry.to_string()
    geometry.multipole = None

    geometry.esp_constraint = (1, 2, 3, 4)
    with pytest.raises(
        InvalidGeometryError,
        match="The provided esp_constraint value is invalid"
    ):
        _ = geometry.to_string()
    geometry.esp_constraint = None


def test_singular():
    """test the singular function"""
    assert singular("A_atoms") == "A"
    assert singular("Bsses") == "Bss"
    assert singular("Cies") == "Cy"
    assert singular("Des") == "De"
    assert singular("gs") == "g"
    assert singular("Z") == "Z"



"""Define the InputSetGenerators for FHI-aims magnetism calculations."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from pymatgen.io.aims.sets.core import RelaxSetGenerator, StaticSetGenerator

if TYPE_CHECKING:
    from typing import Any

    from pymatgen.core.structure import Molecule, Structure


@dataclass
class MagneticStaticSetGenerator(StaticSetGenerator):
    """Common class for ground-state generators.

    Attributes:
        calc_type (str): The type of calculation
    """

    calc_type: str = "static"

    def get_parameter_updates(self, structure: Structure | Molecule, prev_parameters: dict[str, Any]) -> dict[str, Any]:
        """Get the parameter updates for the calculation.

        Args:
            structure (Structure or Molecule): The structure to calculate the bands for
            prev_parameters (Dict[str, Any]): The previous parameters

        Returns:
            dict: The updated for the parameters for the output section of FHI-aims
        """
        updates = {
            "spin": "collinear",
            "output": [*prev_parameters.get("output", []), "mulliken"],
        }
        prev_parameters.update(updates)
        return prev_parameters


@dataclass
class MagneticRelaxSetGenerator(RelaxSetGenerator):
    """Generate FHI-aims relax sets for optimizing internal coordinates and lattice params.

    Attributes:
        calc_type (str): The type of calculation
        relax_cell (bool): If True then relax the unit cell from the structure
        max_force (float): Maximum allowed force in the calculation
        method (str): Method used for the geometry optimization
    """

    def get_parameter_updates(self, structure: Structure | Molecule, prev_parameters: dict[str, Any]) -> dict:
        """Get the parameter updates for the calculation.

        Args:
            structure (Structure or Molecule): The structure to calculate the bands for
        prev_parameters (Dict[str, Any]): The previous parameters

        Returns:
            dict: The updated for the parameters for the output section of FHI-aims
        """
        prev_parameters = super().get_parameter_updates(structure=structure, prev_parameters=prev_parameters)
        updates = {
            "spin": "collinear",
            "output": [*prev_parameters.get("output", []), "mulliken"],
        }
        prev_parameters.update(updates)
        return prev_parameters



"""Utilities for the FHI-aims parser."""

import collections.abc


class NamedDict(dict):
    """A named dict to access items as attributes."""

    __getattr__ = dict.__getitem__


def update(d: dict, u: dict, key: str) -> dict:
    """Update a dictionary `d` with the dictionary `u` under the `key`.

    Parameters
    ----------
    d: dict
        A dictionary to update
    u: dict
        An update dictionary
    key: str
        A key to the update dictionary

    """
    if not key:
        d.update(u)
        return d
    steps = key.split(".")
    step = steps[0]
    # the square brackets only present in the last step of the key
    if step.endswith("[]"):
        # append a list
        d[step[:-2]] = [*d.get(step[:-2], []), u]
    else:
        d_step = d.get(step, {})
        if isinstance(d_step, collections.abc.Mapping):
            d[step] = update(d_step, u, ".".join(steps[1:]))
        else:
            # d_step has to be a list; we update a last dict in it
            assert isinstance(d_step, collections.abc.Iterable)
            d[step][-1] = update(d_step[-1], u, ".".join(steps[1:]))
    return d
